#include <stdio.h>
#include <kallisti/abi/fs.h>
#include "osd_depend.h"

/* optimal random number function. */
unsigned long seed=123;
#define RNDC 1013904223
#define RNDM 1164525

int errno = 0;
void srand(int s)
{
	seed = s;
}

int rand(void) 
{
//	printf("rand\n");
	seed = seed * RNDM + RNDC;
	return seed;
}

int uclock(void)
{
//	printf("uclock\n");
	return thd->jiffies();
}

char tolower(char c) 
{
	return ((c) >= 'A' && (c) <='Z' ? (c)+('a' - 'A') : (c));
}

char toupper(char c) 
{
	return ((c) >= 'a' && (c) <='z' ? (c)-('a' - 'A') : (c));
}

int isupper(char c) 
{
	return ((c) >= 'A' && (c) <='Z' ? 1 : 0);
}

int isxdigit(char c) 
{
	return (((c) >= '0' && (c) <='9') 
		|| ((c) >= 'a' && (c) <='f') 
		|| ((c) >= 'A' && (c) <='F') ? 1 : 0);
}

int isalpha(char c) 
{
	return (((c) >= 'a' && (c) <='z') 
		|| ((c) >= 'A' && (c) <='Z') ? 1 : 0);
}

int isdigit(char c) 
{
	return (((c) >= '0' && (c) <='9') ? 1 : 0);
}

int isspace(char c) 
{
	return ((c) == ' ' ? 1 : 0);
}

int isalnum(char c) 
{
	return (((c) >= '0' && (c) <='9') 
		|| ((c) >= 'a' && (c) <='z') 
		|| ((c) >= 'A' && (c) <='Z') ? 1 : 0);
}

int ispunct(char c) 
{
	return !(isalnum(c) || isspace(c));
}

char *strdup(const char *s)
{
	char * p;
	
	p = malloc(strlen(s) + 1);
	strcpy(p, s);
	return p;
}

char *strstr(const char *searchee, const char *lookfor)
{
//	printf("strstr\n");
  if (*searchee == 0)
    {
      if (*lookfor)
	return (char *) NULL;
      return (char *) searchee;
    }

  while (*searchee)
    {
      int i;
      i = 0;

      while (1)
	{
	  if (lookfor[i] == 0)
	    {
	      return (char *) searchee;
	    }

	  if (lookfor[i] != searchee[i])
	    {
	      break;
	    }
	  i++;
	}
      searchee++;
    }

  return (char *) NULL;
}

void rewind(void) 
{
	printf("rewind\n");
	while (1)
		;
}

void sscanf(void) 
{
	printf("sscanf\n");
	while (1)
		;
}

void vprintf(void) 
{
	printf("vprintf\n");
	while (1)
		;
}

void vfprintf(void) 
{
	printf("vprintf\n");
	while (1)
		;
}

void abort(void) 
{
	printf("abort\n");
	while (1)
		;
}

void fdopen(void) 
{
	printf("fdopen\n");
	while (1)
		;
}

void raise(int signal)
{
	printf("raise\n");
	while (1)
		;
}

void * realloc(void * ptr, size_t size)
{
	void * newptr;
	
//	printf("realloc %d\n", size);
	newptr = malloc(size);
	
	memcpy(newptr, ptr, size);
	
	free(ptr);
	
	return newptr;
}

void * calloc(size_t n, size_t size)
{
	void * newptr;
	
//	printf("calloc %d %d\n", size, n);
	newptr = malloc(n*size);
	memset(newptr, 0, n*size);
	return newptr;
}

#if 0
void assert(int level)
{
	if (level == 0)
	{
		printf("assert %d\n", level);
	
		while (level)
			;
	}
}
#endif

int abs(int number)
{
//	printf("abs\n");
	if (number < 0)
		return -number;
	else
		return number;
}

char fgetc(FILE *stream)
{
	char c;
	
	if (fs->read((unsigned int)stream, &c, 1) == 1)
		return c;
	else
		return EOF;
}

int chdir(const char * d)
{
	printf("chdir %s\n", d);
	while (1)
		;
}

void fputc(void)
{
	printf("fputc\n");
	while (1)
		;
}

void ferror(void)
{
	printf("ferror\n");
	while (1)
		;
}

void fflush(void)
{
	printf("fflush\n");
	while (1)
		;
}

void getchar(void)
{
	printf("getchar\n");
	while (1)
		;
}

void ungetc(void)
{
	printf("ungetc\n");
	while (1)
		;
}

FILE * fopen(const char *path, const char *mode)
{
//	printf("fopen\n");
    if (mode[0] == 'r')
	return (FILE *)fs->open(path, O_RDONLY);
    if (mode[0] == 'w')
	return (FILE *)fs->open(path, O_WRONLY);
	return NULL;
}

#if 0
FILE * open(const char *path, const char *mode)
{
    if (mode[0] == 'r')
	return (FILE *)fs->open(path, O_RDONLY);
    if (mode[0] == 'w')
	return (FILE *)fs->open(path, O_WRONLY);
}
#endif

int fprintf(FILE *stream, const char *fmt, ...)
{
//	printf("fprintf\n");
    va_list args;
    int i;
    char buf[1024];

    if ((int)stream <= 2) {
	va_start(args, fmt);
	i = vsprintf(buf, fmt, args);
	va_end(args);
	
	printf("%s", buf);
	
	return i;
    } else
	return -1;
}

int fclose(FILE *stream)
{
//	printf("fclose\n");
    fs->close((unsigned int)stream);
    return 0;
}

#if 0
int close(FILE *stream)
{
    fs->close((unsigned int)stream);
    return 0;
}
#endif

unsigned int fread(void *ptr, unsigned int size, unsigned int nmemb, FILE *stream)
{
//	printf("fread\n");
    return fs->read((unsigned int)stream, ptr, size*nmemb)/size;
}

unsigned int read(void *ptr, unsigned int size, unsigned int nmemb, FILE *stream)
{
//	printf("read\n");
    return fs->read((unsigned int)stream, ptr, size*nmemb)/size;
}

unsigned int fwrite(void *ptr, unsigned int size, unsigned int nmemb, FILE *stream)
{
//	printf("fwrite\n");
    return fs->write((unsigned int)stream, ptr, size*nmemb)/size;
}

unsigned int write(void *ptr, unsigned int size, unsigned int nmemb, FILE *stream)
{
//	printf("write\n");
    return fs->write((unsigned int)stream, ptr, size*nmemb)/size;
}

int fseek(FILE *stream, long offset, int whence)
{
//	printf("fseek %d, %d\n",offset,whence);
    if (fs->seek((unsigned int)stream, offset, whence) == -1)
    	return 1;
    
    return 0;
}

int seek(FILE *stream, long offset, int whence)
{
//	printf("seek\n");
    return fs->seek((unsigned int)stream, offset, whence);
}

int fpos(FILE *stream)
{
//	printf("fpos\n");
    return fs->tell((unsigned int)stream);
}

int ftell(FILE *stream)
{
//	printf("ftell\n");
    return fs->tell((unsigned int)stream);
}

int feof(FILE *stream)
{
//	printf("feof\n");
	if (fs->total((unsigned int)stream) == ftell(stream))
		return 0;
	else
		return -1; 
}

int pos(FILE *stream)
{
//	printf("pos\n");
    return fs->tell((unsigned int)stream);
}

int fstat(FILE *stream, struct stat * fileinfo)
{
//	printf("fstat\n");
	fileinfo->st_size = fs->total((unsigned int)stream);
    return fileinfo->st_size;
}

int stat(char *fname, struct stat * fileinfo)
{
	FILE * fd;

	fd = fopen(fname,"rb");
	
	if (fd)
	{
//		printf ("stat %s, fd = %d\n", fname, fd);

		fileinfo->st_size = fs->total((unsigned int)fd);
		fileinfo->st_mode = S_IFDIR;
		fclose((void*)fd);
		return 0;
	}
    
//	printf ("stat not found %s\n", fname);

    return 1;
}

int access(const char * fname, int permissions)
{
	FILE * fd;
	
	fd = fopen(fname,"rb");
	
	if (fd)
	{
//		printf("access found %s %d\n", fname, fd);	
		fclose(fd);
		return 0;
	}

//	printf("access not found %s\n",fname);	
	return -1;
}

#if 0
int strncasecmp(const char *inString1, const char *inString2, unsigned int length)
{
    char c1, c2;
    int walk;
    for (walk = 0; walk < length; walk++)
    {
    	if (isupper(*inString1))
			c1 = tolower(*inString1);
		else
			c1 = *inString1;
		inString1++;
    	if (isupper(*inString2))
    		c2 = tolower(*inString2);
    	else
    		c2 = *inString2;
    	inString2++;
        if (c1 < c2) return -1;
        if (c1 > c2) return 1;
        if (c1 == 0) return 0;
    }
    
    return 0;
}

int strcasecmp(const char *inString1, const char *inString2)
{
    char c1, c2;
    while (1)
    {
    	if (isupper(*inString1))
			c1 = tolower(*inString1);
		else
			c1 = *inString1;
		inString1++;
    	if (isupper(*inString2))
    		c2 = tolower(*inString2);
    	else
    		c2 = *inString2;
    	inString2++;
        if (c1 < c2) return -1;
        if (c1 > c2) return 1;
        if (c1 == 0) return 0;
    }
}
#endif
void exit(int status)
{
    printf("fake exit() called - you must reset now\n");    
    while(1)
    	;
}

void __extendsfdf2(void)
{
	printf("__extendsfdf2\n");
	while (1)
		;
}

void __truncdfsf2(void)
{
	printf("__extendsfdf2\n");
	while (1)
		;
}

char * fgets(char *buf, int n, FILE * fp)
{
	char *s;
	int c = EOF;

	if (n < 2)			/* sanity check */
		return 0;

	s = buf;

	/* Sorry, have to do it the slow way */
	while (--n > 0 && (c = fgetc(fp)) != EOF)
	{
		*s++ = c;
		if (c == '\n')
			break;
	}
	if (c == EOF && s == buf)
	{
		return NULL;
	}
	
	*s = 0;
	
	return buf;
}

#ifndef ULONG_MAX
#define	ULONG_MAX	((unsigned long)(~0L))		/* 0xFFFFFFFF */
#endif

#ifndef LONG_MAX
#define	LONG_MAX	((long)(ULONG_MAX >> 1))	/* 0x7FFFFFFF */
#endif

#ifndef LONG_MIN
#define	LONG_MIN	((long)(~LONG_MAX))		/* 0x80000000 */
#endif

long
strtol(nptr, endptr, base)
	const char *nptr;
	char **endptr;
	register int base;
{
	register const char *s = nptr;
	register unsigned long acc;
	register int c;
	register unsigned long cutoff;
	register int neg = 0, any, cutlim;

	/*
	 * Skip white space and pick up leading +/- sign if any.
	 * If base is 0, allow 0x for hex and 0 for octal, else
	 * assume decimal; if base is already 16, allow 0x.
	 */
	do {
		c = *s++;
	} while (isspace(c));
	if (c == '-') {
		neg = 1;
		c = *s++;
	} else if (c == '+')
		c = *s++;
	if ((base == 0 || base == 16) &&
	    c == '0' && (*s == 'x' || *s == 'X')) {
		c = s[1];
		s += 2;
		base = 16;
	}
	if (base == 0)
		base = c == '0' ? 8 : 10;

	/*
	 * Compute the cutoff value between legal numbers and illegal
	 * numbers.  That is the largest legal value, divided by the
	 * base.  An input number that is greater than this value, if
	 * followed by a legal input character, is too big.  One that
	 * is equal to this value may be valid or not; the limit
	 * between valid and invalid numbers is then based on the last
	 * digit.  For instance, if the range for longs is
	 * [-2147483648..2147483647] and the input base is 10,
	 * cutoff will be set to 214748364 and cutlim to either
	 * 7 (neg==0) or 8 (neg==1), meaning that if we have accumulated
	 * a value > 214748364, or equal but the next digit is > 7 (or 8),
	 * the number is too big, and we will return a range error.
	 *
	 * Set any if any `digits' consumed; make it negative to indicate
	 * overflow.
	 */
	cutoff = neg ? -(unsigned long)LONG_MIN : LONG_MAX;
	cutlim = cutoff % (unsigned long)base;
	cutoff /= (unsigned long)base;
	for (acc = 0, any = 0;; c = *s++) {
		if (isdigit(c))
			c -= '0';
		else if (isalpha(c))
			c -= isupper(c) ? 'A' - 10 : 'a' - 10;
		else
			break;
		if (c >= base)
			break;
		if ((((any < 0) || (acc > cutoff) || (acc == cutoff)) && (c)) > cutlim)
			any = -1;
		else {
			any = 1;
			acc *= base;
			acc += c;
		}
	}
	if (any < 0) {
		acc = neg ? LONG_MIN : LONG_MAX;
//		errno = -1;
	} else if (neg)
		acc = -acc;
	if (endptr != 0)
		*endptr = (char *) (any ? s - 1 : nptr);
	return (acc);
}

int atoi( char *s)
{
	printf("atoi\n");
  	return (int) strtol (s, NULL, 10);
}

