// standard OS includes (vxWorks)
//#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
//#include <sys/stat.h>
//#include <dirent.h>

// MAME includes
#include "osdepend.h"
#include "osd_depend.h"
#include "driver.h"

// local prototypes
#ifdef MESS
void messfront(void);
#else
void mamefront(void);
#endif

// external prototypes
extern void (*update_screen)(struct osd_bitmap *bitmap);
extern void* kos_audio_callback(int size);
extern int kos_stream_init(void* (*callback)(int));

#define kREPEAT_TIME 7

// external globals
extern int					use_dirty;
extern struct snd_interface * sndintf;

extern void undefined_address(void);
extern void stream_shutdown(void);
extern void set_config_file(char * fname);
extern void parse_cmdline (int game_index);
extern void * mame (char * gamename);
extern UINT8 *create_15bit_palette ( void );
#include "driver.h"
#include "artwork.h"
#include "png.h"
struct artwork_info *snap_overlay = NULL;

extern void allocate_artwork_mem (int width, int height, struct artwork_info **a);



// Globals
unsigned long 				*gLCD_buffer1 = 0;
unsigned long 				*gLCD_buffer2 = 0;
int 						gResolution = 0;
int 						gPadOrientation = 0;
int							play_sound = 0;
int							cpu_slices_per_frame_override = 0;
int 						auto_rotate = 0;
int 						saveUse_dirty = -1;
int 						hide_clones = 0;

/* Library handles */
abi_fs_t    	*fs 	= NULL;
abi_thread_t    *thread = NULL;
abi_video_t 	*video 	= NULL;
abi_maple_t 	*maple 	= NULL;
abi_conio_t 	*conio 	= NULL;
abi_thread_t 	*thd 	= NULL;
abi_spu_t		*spu	= NULL;
abi_ta_t 		*ta 	= NULL;
abi_fundamental_t *fabi = NULL;

static	void*	(*oldmalloc)(size_t size) = NULL;
static	void	(*oldfree)(void *ptr) = NULL;
static	void*	(*oldmemalign)(size_t size, size_t align) = NULL;

static	void*	mamed_malloc(size_t size);
static	void	mamed_free(char *ptr);
static	void*	mamed_memalign(size_t size, size_t align);
static void init_memory_table(void);
static void shutdown_memory_table(void);

int max_memory_in_use = 0;
int total_memory_in_use = 0;
int current_malloc_count =  0;

#define MAX_MEMORY_TABLE 10000
#define HEAD_FLAG 0x12345678UL
#define TAIL_FLAG 0x87654321UL

struct
{
	char *address;
	size_t size;
	int in_use;
	int count;
} memory_table[MAX_MEMORY_TABLE];

/* empty list that will be replaced with _drivers[] when we load the game modules */
extern const struct GameDriver *_drivers[];
const struct GameDriver **drivers = _drivers;

/* The "root" driver, defined so we can have &driver_##NAME in macros. */
#ifdef MESS
struct GameDriver driver_0 =
{
  __FILE__,
  0,
  "root",
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  NOT_A_DRIVER,
};
#else
struct GameDriver driver_0 =
{
	__FILE__,
	0,
	"",
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	NOT_A_DRIVER
};
#endif

void fix_driver_table(void)
{
	struct GameDriver **walk_drivers_link;
	struct GameDriver **set_drivers_link;

	// remove gaps (zeros or unresolved references) in the game driver list,
	// cause MAME expects it that way.
	walk_drivers_link = (void *)drivers;
	set_drivers_link = (void *)drivers;
	while (*walk_drivers_link != (struct GameDriver *)0xFFEEDDCC)
	{
		while ((*walk_drivers_link == 0) || (*walk_drivers_link == (void *)undefined_address))
			walk_drivers_link++;
		
		if (*walk_drivers_link != (struct GameDriver *)0xFFEEDDCC)
		{
			*set_drivers_link = *walk_drivers_link;
			set_drivers_link++;
			walk_drivers_link++;
		}
	}
	*set_drivers_link = 0;
}

static char * vmu_icon =
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................"
"................................................";

void vmu_icon_init() 
{
    int x, y, xi, xb;
    uint8 bitmap[48*32/8];
    uint8 addr;

    addr = maple->first_vmu();
    if (!addr) return;

    memset(bitmap, 0, 48*32/8);
    for (y=0; y<32; y++)
	for (x=0; x<48; x++) {
	    xi = x / 8;
	    xb = 0x80 >> (x % 8);
	    if (vmu_icon[(31-y)*48+(47-x)] == '.')
		bitmap[y*(48/8)+xi] |= xb;
	}
    maple->vmu_draw_lcd(addr, bitmap);
}

int save_argc;
char ** save_argv;

int main(int argc, char *argv[])
{
	int i;
	
	save_argc = argc;
	save_argv = argv;
	if ((argc != 2) && (argc != 1))
	{
		printf("Usage mamed <romname>\n\r");
	}

	fix_driver_table();

	fabi = lib_open("fundamental");
	if(!fabi)
	{
		printf("Can't open fabi lib\n");
		return 1;
	}

	thd = lib_open("thread");
	if (!thd) 
	{ 
		printf("Can't open thread library\r\n"); 
		return 1; 
	}

    fs = lib_open("fs");
    if (fs == NULL) 
    { 
    	printf("Can't open fs lib\r\n"); 
    	return 1; 
    }

    video = lib_open("video");
    if (video == NULL) 
    { 
    	printf("Can't open video lib\r\n"); 
    	return 1; 
    }

    maple = lib_open("maple");
    if (maple == NULL) 
    { 
    	printf("Can't open maple lib\r\n"); 
    	return 1; 
    }

	ta = lib_open("ta");
	if (ta == NULL) 
	{ 
		printf("Can't open ta lib\r\n"); 
		return 1; 
	}

	spu = lib_open("spu");
	if (spu == NULL) 
	{
		printf("Can't open spu!\r\n");
		return 1;
	}

	conio = lib_open("conio");
	if (conio != NULL) 
	{
		printf("Kosh detected; freezing\r\n");
		conio->freeze();
	}
	
	vmu_icon_init();

	// setup the stream audio data callback
	kos_stream_init(kos_audio_callback);

	init_memory_table();

	// start  MAME thread
#ifdef MESS
		messfront(); 
#else
		mamefront(); 
#endif

	// OK, nasty HACK time. Everything should be freed in memory right now
	// But it's not, so I will manually free everything in the list to prevent
	// memory fragmentation.

	printf("HACK!!! clean it up\n");
	for (i = 0; i < MAX_MEMORY_TABLE; i++)
	{
		if (memory_table[i].in_use == 1)
		{
			free(memory_table[i].address + 4);
		}
	}
	printf("done cleaning it up\n");

	shutdown_memory_table();
	
	stream_shutdown();
	
	video->set_mode(video->check_cable(), DM_640x480, PM_RGB565);
	video->clear(0,0,0);

	if (conio != NULL) 
	{
		printf("Kosh detected; thawing\r\n");
		conio->thaw();
	}
#if 0
	lib_close(conio);
	lib_close(spu);
	lib_close(ta);
	lib_close(maple);
	lib_close(video);
	lib_close(fs);
	lib_close(thd);
#endif

	return 0;
}

#define CPU_TYPE(driver_index, index) (drivers[driver_index]->drv->cpu[index].cpu_type & ~CPU_FLAGS_MASK)
#define CPU_AUDIO(driver_index, index) (drivers[driver_index]->drv->sound[index].sound_type & CPU_AUDIO_CPU)

// note: this doesn't completely verify the game will actually run, but it's a start
// note: doesn't work on dc cause dynamic loaded unresolved references are not zero
int Flight_Check(int driver_index)
{
	int totalcpu = 0;
	int totalsound = 0;
	
	while (totalcpu < MAX_CPU)
	{
		int cpu_type = CPU_TYPE(driver_index, totalcpu);
		
		if(cpu_type  == CPU_DUMMY ) break; // end of list
		
		if ((cpuintf[cpu_type].execute == 0) || ((void *)cpuintf[cpu_type].execute == (void *)undefined_address))
			return -1; // this cpu is not loaded, error
			
		totalcpu++;
	}
	
	while (totalsound < MAX_SOUND)
	{
		int sound_type = CPU_AUDIO(driver_index, totalsound);
		
		if(sound_type  == 0 ) break; // end of list

		if ((sndintf[sound_type].start == 0 ) || ((void *)cpuintf[sound_type].execute == (void *)undefined_address))
			return -1; // this sound driver is not loaded, error

		totalsound++;
	}

	return 0; // good to go
}

#define CPU_COUNT 49

void Excess_And_Missing_Check(void)
{
	int i;
	int cpu_used_table[CPU_COUNT];
	int cpu_missing_table[CPU_COUNT];
	
	for (i = 0; i < CPU_COUNT; i++)
	{
		cpu_used_table[i] = 0;
		cpu_missing_table[i] = 0;
	}
	
	for (i = 0; drivers[i]; i++)
	{
		if (drivers[i])
		{
			int totalcpu = 0;

			while (totalcpu < MAX_CPU)
			{
				int cpu_type = CPU_TYPE(i, totalcpu);
				
				if(cpu_type  == CPU_DUMMY ) break; // end of list
				
				if ((cpuintf[cpu_type].execute == 0) || ((void *)cpuintf[cpu_type].execute == (void *)undefined_address))
				{
					cpu_missing_table[cpu_type]++; // this cpu is not loaded, error
				}
				else
				{
					cpu_used_table[cpu_type]++;
				}
				
				totalcpu++;
			}
		}
	}

	for (i = 0; i < CPU_COUNT; i++)
	{
		if (cpu_used_table[i] != 0)
		{
			printf("Used CPU #%d by %d drivers %s \n", i, cpu_used_table[i], cputype_name(i));
		}
		
		if (cpu_missing_table[i] != 0)
		{
			printf("Missing CPU #%d, used by %d drivers\n", i, cpu_missing_table[i]);
		}
		
		if ((cpu_missing_table[i] == 0) && (cpu_used_table[i] == 0))
		{
			if ((cpuintf[i].execute == 0) || ((void *)cpuintf[i].execute == (void *)undefined_address))
				;
			else
				printf("Excess CPU #%d %s\n", i, cputype_name(i));
		}
	}
}

int sort_func(const char **a, const char **b)
{
	return stricmp(*a, *b);
}

// Taken from artwork.c and modified to load snaps
/* static */ int snap_read_bitmap(const char *file_name, struct osd_bitmap **bitmap, struct osd_bitmap **alpha, struct png_info *p)
{
	UINT8 *tmp;
	int x, y, pen;
	void *fp;
	int file_name_len;
	char file_name2[256];

	/* check for .png */
	strcpy(file_name2, file_name);
	file_name_len = strlen(file_name2);
	if ((file_name_len < 4) || stricmp(&file_name2[file_name_len - 4], ".png"))
	{
		strcat(file_name2, ".png");
	}

	if (!(fp = osd_fopen(Machine->gamedrv->name, file_name2, OSD_FILETYPE_SCREENSHOT, 0)))
	{
		return 0;
	}

	if (!png_read_file(fp, p))
	{
		osd_fclose (fp);
		return 0;
	}
	osd_fclose (fp);

	if (p->bit_depth > 8)
	{
		printf("Unsupported bit depth %i (8 bit max.)\n", p->bit_depth);
		return 0;
	}

	if (p->interlace_method != 0)
	{
		printf("Interlace unsupported\n");
		return 0;
	}

	if (Machine->scrbitmap->depth == 8 && p->color_type != 3)
	{
		printf("Use 8bit artwork for 8bpp modes. Artwork disabled.\n");
		return 0;
	}

	switch (p->color_type)
	{
	case 3:
		/* Convert to 8 bit */
		png_expand_buffer_8bit (p);
		png_delete_unused_colors (p);

		if ((*bitmap = bitmap_alloc(p->width,p->height)) == 0)
		{
			printf("Unable to allocate memory for artwork\n");
			return 0;
		}

		tmp = p->image;
		if ((*bitmap)->depth == 8)
		{
			for (y=0; y<p->height; y++)
				for (x=0; x<p->width; x++)
				{
					plot_pixel(*bitmap, x, y, *tmp++);
				}
		}
		else
		{
			/* convert to 15bit */
			if (p->num_trans > 0)
				if ((*alpha = bitmap_alloc(p->width,p->height)) == 0)
				{
					printf("Unable to allocate memory for artwork\n");
					return 0;
				}

			for (y=0; y<p->height; y++)
				for (x=0; x<p->width; x++)
				{
//					pen = ((p->palette[*tmp * 3] & 0xf8) << 7) | ((p->palette[*tmp * 3 + 1] & 0xf8) << 2) | (p->palette[*tmp * 3 + 2] >> 3);
					pen = ((p->palette[*tmp * 3] & 0xf8) << 8) | ((p->palette[*tmp * 3 + 1] & 0xf8) << 3) | (p->palette[*tmp * 3 + 2] >> 3);
					plot_pixel(*bitmap, x, y, pen);

					if (p->num_trans > 0)
					{
						if (*tmp < p->num_trans)
							plot_pixel(*alpha, x, y, p->trans[*tmp]);
						else
							plot_pixel(*alpha, x, y, 255);
					}
					tmp++;
				}

			free (p->palette);

			/* create 15 bit palette */
			if ((p->palette = (UINT8 *)create_15bit_palette()) == 0)
			{
				printf("Unable to allocate memory for artwork\n");
				return 0;
			}
			p->num_palette = 32768;
		}
		break;

	case 6:
		if ((*alpha = bitmap_alloc(p->width,p->height)) == 0)
		{
			printf("Unable to allocate memory for artwork\n");
			return 0;
		}

	case 2:
		if ((*bitmap = bitmap_alloc(p->width,p->height)) == 0)
		{
			printf("Unable to allocate memory for artwork\n");
			return 0;
		}

		/* create 15 bit palette */
		if ((p->palette = (UINT8 *)create_15bit_palette()) == 0)
		{
			printf("Unable to allocate memory for artwork\n");
			return 0;
		}

		p->num_palette = 32768;
		p->trans = NULL;
		p->num_trans = 0;

		/* reduce true color to 15 bit */
		tmp = p->image;
		for (y=0; y<p->height; y++)
			for (x=0; x<p->width; x++)
			{
//				pen = ((tmp[0] & 0xf8) << 7) | ((tmp[1] & 0xf8) << 2) | (tmp[2] >> 3);
				pen = ((tmp[0] & 0xf8) << 8) | ((tmp[1] & 0xf8) << 3) | (tmp[2] >> 3);
				plot_pixel(*bitmap, x, y, pen);

				if (p->color_type == 6)
				{
					plot_pixel(*alpha, x, y, tmp[3]);
					tmp += 4;
				}
				else
					tmp += 3;
			}

		break;

	default:
		printf("Unsupported color type %i \n", p->color_type);
		return 0;
		break;
	}
	free (p->image);
	return 1;
}

// Taken from artwork.c and modified to load snaps
/* static */ void snap_load_size_common(const char *filename, unsigned int start_pen, unsigned int max_pens,
					   				 int width, int height, struct artwork_info **a, int * orig_width, int * orig_height)
{
	struct osd_bitmap *picture = 0, *alpha = 0;
	struct png_info p;
	int scalex, scaley;

	/* If the user turned artwork off, bail */
//	if (!options.use_artwork) return;

	allocate_artwork_mem(width, height, a);

	if (*a==NULL)
		return;

	(*a)->start_pen = start_pen;

	if (!snap_read_bitmap(filename, &picture, &alpha, &p))
	{
		artwork_free(a);
		return;
	}

	(*a)->num_pens_used = p.num_palette;
	(*a)->num_pens_trans = p.num_trans;
	(*a)->orig_palette = p.palette;
	(*a)->transparency = p.trans;

	/* Make sure we don't have too many colors */

	if ((*a)->num_pens_used > max_pens)
	{
		printf("Too many colors in artwork.\n");
		printf("Colors found: %d  Max Allowed: %d\n",
				      (*a)->num_pens_used,max_pens);
		artwork_free(a);
		bitmap_free(picture);
		return;
	}

	/* Scale the original picture to be the same size as the visible area */
	scalex = 0x10000 * picture->width  / width;
	scaley = 0x10000 * picture->height / height;
	if (Machine->orientation & ORIENTATION_SWAP_XY)
	{
		int tmp;
		tmp = scalex;
		scalex = scaley;
		scaley = tmp;
	}

	if (scalex < scaley)
		scalex = scaley;
	if (scaley < scalex)
		scaley = scalex;
	copyrozbitmap((*a)->orig_artwork, picture, 0, 0, scalex, 0, 0, scaley, 0, 0, TRANSPARENCY_NONE, 0, 0);

	*orig_width =  0x10000 * picture->width / scalex;
	*orig_height = 0x10000 * picture->height / scaley ;

	/* We don't need the original any more */
	bitmap_free(picture);

	if (alpha)
	{
		copyrozbitmap((*a)->alpha, alpha, 0, 0, scalex, 0, 0, scaley, 0, 0, TRANSPARENCY_NONE, 0, 0);
		bitmap_free(alpha);
	}

	/* If the game uses dynamic colors, we assume that it's safe
	   to init the palette and remap the colors now */
//	if (Machine->drv->video_attributes & VIDEO_MODIFIES_PALETTE)
//		backdrop_set_palette(*a,(*a)->orig_palette);
}

#define NUMBER_OF_MENU_COLORS 2
UINT8 dummy_palette[NUMBER_OF_MENU_COLORS][3] = {{0,0,0}, {255,255,255}};
UINT16 pens[NUMBER_OF_MENU_COLORS];

#ifdef MESS
int Validate_ROM(char * filename, const struct IODevice *dev)
{
	char extension[10];
	char * ext;
	
	/* find extension */
	ext = strrchr(filename, '.');
	if (ext)
	{
		ext++;
		strcpy(extension, ext);

		// delete spaces any from extension
		if (extension[0] == ' ')
		{
			extension[0] = 0;
		}
		else if (extension[1] == ' ')
		{
			extension[1] = 0;
		}
		else if (extension[2] == ' ')
		{
			extension[2] = 0;
		}

		while (dev->type != IO_END)
		{
			const char *dst = dev->file_extensions;
			/* scan supported extensions for this device */
			while (dst && *dst)
			{
				// assume first file type for zipped files
				if (stricmp("zip",extension) == 0)
				{
					return dev->type;
				}

				if (stricmp(dst,extension) == 0)
				{
                    return dev->type;
				}
				/* skip '\0' once in the list of extensions */
				dst += strlen(dst) + 1;
			}
			dev++;
		}
	}
	
	return -1;
}
#if 0
#define DIR dirent_t
int Select_ROM(char *dirpath, const struct IODevice *dev, char * rom_name)
{
    DIR     		*b;
    dirent_t 		*c;
	int 			sel;
	int 			menu_total;
	char 			**menu_item;
	int 			visible;
	int 			rom_type;
		
	// clear the screen
	osd_clearbitmap(Machine->scrbitmap);

	// first is default selected
	sel = 0;

	visible = (Machine->uiheight / (3 * Machine->uifontheight / 2))/2 - 1;

	// count number of files in rom directory
	menu_total = 0;
    b = open(dirpath);
    if (b) 
    {
		while ((c = readdir(b)))
		{
			if (Validate_ROM(c->d_name, dev) != -1)
			{
				menu_total++;
			}
		}
		close(b);
	}

	// allocate enough space
	menu_item = (char **)malloc(sizeof(char **) * (menu_total + 2));

	menu_total = 0;
	
	// add first one for blank selection
	menu_item[menu_total++] = "";
	
    b = open(dirpath);
    if (b) 
    {
		while ((c = readdir(b))) 
		{
			rom_type = Validate_ROM(c->d_name, dev);
			if (rom_type != -1)
			{
				menu_item[menu_total] = malloc(strlen(c->d_name) + 1);
				strcpy(menu_item[menu_total], c->d_name);
				menu_total++;
			}
		}
		close(b);
	}

	// end of the list
	menu_item[menu_total] = 0;

	while (TRUE)
	{
		ui_displaymenu(Machine->scrbitmap, (const char **)menu_item,0,0,sel,0);
		update_screen(Machine->scrbitmap);

		if (input_ui_pressed_repeat(IPT_UI_DOWN,kREPEAT_TIME))
			sel = sel + 1;

		if (input_ui_pressed_repeat(IPT_UI_UP,kREPEAT_TIME))
			sel = sel - 1;
			
		if (input_ui_pressed_repeat(IPT_UI_RIGHT,kREPEAT_TIME))
			sel = sel + visible - 1;

		if (input_ui_pressed_repeat(IPT_UI_LEFT,kREPEAT_TIME))
			sel = sel - visible + 1;

		if (sel > menu_total - 1)
		{
			sel = 0;
		}
		else if (sel < 0)
		{
			sel = menu_total - 1;
		}
		
		if (input_ui_pressed(IPT_UI_SELECT) || input_ui_pressed(IPT_START1))
		{
			break;
		}
		
		if (input_ui_pressed(IPT_UI_CANCEL))
		{
			sel = -1;
			break;
		}
	}

	if (sel != -1)
	{
		// save a copy in return string
		strcpy(rom_name, menu_item[sel]);
	}
	
	// free all the strings
	// skip the first one since it's hard coded
	menu_total = 1;
	while(menu_item[menu_total])
	{
		free(menu_item[menu_total]);
		menu_total++;
	}

	// free the table
	free(menu_item);
	
	if (sel != -1)
	{
		return Validate_ROM(rom_name, dev);
	}
	else
	{
		return -2;
	}
}
#endif
char rom_files[MAX_IMAGES][20];

void messfront(void)
{
	int sel;
	int visible;
	int hide_menu = 0;
	int lastone = -9999;
	int menu_total;
	char **menu_item;
	int i;
	int index;

	memset(&options, 0, sizeof(options));
	// Init default options
	options.norotate     = 0;
	options.ror          = 0;
	options.rol          = 0;
	options.flipx        = 0;
	options.flipy        = 0;
	options.use_artwork  = 1;
	options.color_depth  = 16;
	options.image_count  = 0;
	
	options.use_samples  = 1;
	options.samplerate   = 44100;
	play_sound = 1;
	
	use_dirty = 0; // off 

	gResolution = 0;
	gPadOrientation = 0;
	cpu_slices_per_frame_override = 1;
	
    set_config_file ("/rd/MESS.CFG");

	parse_cmdline(0);
	
	// first is default selected
	sel = 0;
	
	// count drivers so we can allocate enough memory
	menu_total = 0;
	for (i = 0;drivers[i];i++)
	{
		if (drivers[i])
		{
			if (isalnum(drivers[i]->name[0]))
			{
				// check if rom exists and add it to list
				if (osd_faccess (drivers[i]->name, OSD_FILETYPE_ROM))
				{
					menu_total++;
				}
			}
		}
	}
	menu_item = (char **)malloc(sizeof(char **) * (menu_total + 1));
	
	// reset count to start again and this time add the ones we find
	menu_total = 0;
	for (i = 0;drivers[i];i++)
	{
		if (drivers[i])
		{
			if (isalnum(drivers[i]->name[0]))
			{
				// check if rom exists and add it to list
				if (osd_faccess (drivers[i]->name, OSD_FILETYPE_ROM))
				{
					if (Flight_Check(i))
					{
						int x;
						char * name;

						// mark file as not running by making it upper case
						for (x = 0; x < strlen(drivers[i]->name); x++)
						{
							name = (char *)drivers[i]->name;
							name[x] = toupper(name[x]);
						}
					}
					menu_item[menu_total++] = (char *)drivers[i]->name;
				}
			}
		}
	}
	
	// end of the list
	menu_item[menu_total] = 0;

//	qsort(menu_item, menu_total, sizeof(char **), sort_func);

	while(TRUE)
	{
		struct GameOptions saveOptions;
		int save_play_sound;
		int saveResolution;
		int savePadOrientation;
		struct MachineDriver fake_driver;
		struct GameDriver fake_gamedrv;
		struct InputPort fake_input_port;

		// reset screen to normal orientation and resolution while in main menu
		// we will restore it just before we start the game		

		memset(&fake_driver, 0, sizeof(fake_driver));
		memset(&fake_gamedrv, 0, sizeof(fake_gamedrv));
		memset(&fake_input_port, 0, sizeof(fake_input_port));
		
		saveOptions = options;
		save_play_sound = play_sound;
		saveResolution = gResolution;
		savePadOrientation = gPadOrientation;
		saveUse_dirty = use_dirty; 
		
		options.norotate    = 0;
		options.ror         = 0;
		options.rol         = 0;
		options.flipx       = 0;
		options.flipy       = 0;
		options.samplerate  = 0;
		options.use_samples = 0;
		options.use_artwork = 0;
		options.color_depth  = 16;
		options.image_count = 0;
		
//		play_sound = 1;
		gResolution = 0;
		gPadOrientation = 0;

		// Init enough of MAME's structures to get the user menus to work		
		
		Machine->drv = &fake_driver;
		fake_driver.frames_per_second = 60;
		fake_driver.video_attributes = 0; // VIDEO_MODIFIES_PALETTE;
		fake_driver.total_colors = NUMBER_OF_MENU_COLORS;
		fake_driver.screen_width = 320;
		fake_driver.screen_height = 240;

		Machine->gamedrv = &fake_gamedrv;
		
		Machine->orientation = ROT0;
		Machine->ui_orientation = ROT0;
		Machine->color_depth = 16;

		Machine->input_ports_default = &fake_input_port;
		Machine->input_ports = &fake_input_port;
		fake_input_port.type = IPT_END;

		code_init();

		Machine->scrbitmap = bitmap_alloc_depth(fake_driver.screen_width,fake_driver.screen_height,Machine->color_depth);
		osd_create_display(fake_driver.screen_width,
			fake_driver.screen_height,
			Machine->color_depth,
			fake_driver.frames_per_second,0,0);
		set_visible_area(0,fake_driver.screen_width,0,fake_driver.screen_height);		
		Machine->uifont = builduifont();

		visible = (Machine->uiheight / (3 * Machine->uifontheight / 2))/2 - 1;

		osd_allocate_colors(Machine->drv->total_colors, (void *)dummy_palette, pens, 0, 0, 0);
		
		for (i = 0; i < 65536; i++);
			pens[i] = i;
			
		while (TRUE)
		{
			if (menu_total == 0) // somethings wrong, nothing in the list, let the user reconfigure and then reboot
			{

			}

#if 1			
			if (input_ui_pressed(IPT_COIN1))
			{
				hide_menu = !hide_menu;
				lastone = -9999;
			}

			// draw snapshot before menu
			if (sel != lastone)
			{
				int orig_width;
				int orig_height;

				lastone = sel;
				
				// save two colors for new UI B&W
				snap_load_size_common(menu_item[sel], 0, 32768, Machine->scrbitmap->width, Machine->scrbitmap->height, &snap_overlay, &orig_width, &orig_height);

				if (snap_overlay == NULL)
				{
					// find the driver in the list
					for (i = 0;drivers[i];i++)
					{
						if (drivers[i])
						{
							if (drivers[i]->name == menu_item[sel])
							{
								break;
							}
						}
					}

					if (drivers[i]->clone_of != &driver_0)
						snap_load_size_common(drivers[i]->clone_of->name, 0, 32768, Machine->scrbitmap->width, Machine->scrbitmap->height, &snap_overlay, &orig_width, &orig_height);
					
					// if it's still NULL
					if (snap_overlay == NULL)
						snap_load_size_common("MAMED", 0, 32768, Machine->scrbitmap->width, Machine->scrbitmap->height, &snap_overlay, &orig_width, &orig_height);
				}
				
				osd_clearbitmap(Machine->scrbitmap);
				if (snap_overlay)
				{
					int offsetx;
					int offsety;
					
					for (i = 0; i < snap_overlay->num_pens_used; i++)
						osd_modify_pen(i, snap_overlay->orig_palette[i * 3], 
							snap_overlay->orig_palette[i * 3 + 1], 
							snap_overlay->orig_palette[i * 3 + 2]);

					offsetx = (Machine->scrbitmap->width - orig_width) / 2;
					offsety = (Machine->scrbitmap->height - orig_height) / 2;
					
					copybitmap(Machine->scrbitmap, snap_overlay->orig_artwork, 0, 0, offsetx, offsety, 0, TRANSPARENCY_NONE, 0);
					osd_mark_dirty(0,0,Machine->scrbitmap->width, Machine->scrbitmap->height);
					
					osd_modify_pen(snap_overlay->num_pens_used, 0,0,0);
					osd_modify_pen(snap_overlay->num_pens_used + 1, 255,255,255);

					Machine->uifont->colortable[0] = snap_overlay->num_pens_used;
					Machine->uifont->colortable[1] = snap_overlay->num_pens_used + 1;
					Machine->uifont->colortable[2] = snap_overlay->num_pens_used + 1;
					Machine->uifont->colortable[3] = snap_overlay->num_pens_used;	

					artwork_free(&snap_overlay);
					snap_overlay = NULL;
				}
				else
				{
					hide_menu = 0;
				}
			}
#endif
			// draw menu
			if (!hide_menu)
			{		
				ui_displaymenu(Machine->scrbitmap, (const char **)menu_item,0,0,sel,0);
			}
			
			// get it all to the screen
			update_screen(Machine->scrbitmap);

			if (input_ui_pressed(IPT_UI_CONFIGURE))
			{
				// my little options menu
				const char *game_submenu_item[20];
				int game_submenu_total;
				const char *game_menu_item[20];
				int game_menu_total;
				int gamesel;
				
				// so we refresh the screen shot when we come back
				lastone = -9999;
				osd_clearbitmap(Machine->scrbitmap);

				gamesel = 0;
				
				while (TRUE)
				{
					game_menu_total = 0;
					game_submenu_total = 0;
					
					game_menu_item[game_menu_total++] = "Pad Orientation"; 
					if (savePadOrientation == 0) 
						game_submenu_item[game_submenu_total++] = "Up";
					else if (savePadOrientation == 1)
						game_submenu_item[game_submenu_total++] = "Right";
					else if (savePadOrientation == 2)
						game_submenu_item[game_submenu_total++] = "Down";
					else if (savePadOrientation == 3)
						game_submenu_item[game_submenu_total++] = "Left";
					else
						game_submenu_item[game_submenu_total++] = "Unknown";

					game_menu_item[game_menu_total++] = "Resolution";
					if (saveResolution == 1) 
						game_submenu_item[game_submenu_total++] = "640x480TAFill";
					else if (saveResolution == -1) 
						game_submenu_item[game_submenu_total++] = "640x480TAScale";
					else if (saveResolution == 0) 
						game_submenu_item[game_submenu_total++] = "320x240Direct";
					else if (saveResolution == 2) 
						game_submenu_item[game_submenu_total++] = "320x240TAFill";
					else if (saveResolution == 3) 
						game_submenu_item[game_submenu_total++] = "640x480Direct";
					
					game_menu_item[game_menu_total++] = "No Rotation";
					if (saveOptions.norotate) 
						game_submenu_item[game_submenu_total++] = "Yes";
					else 
						game_submenu_item[game_submenu_total++] = "No"; 
					game_menu_item[game_menu_total++] = "Rotation Right"; 
					if (saveOptions.ror) 
						game_submenu_item[game_submenu_total++] = "Yes";
					else 
						game_submenu_item[game_submenu_total++] = "No"; 
					game_menu_item[game_menu_total++] = "Rotation Left"; 
					if (saveOptions.rol) 
						game_submenu_item[game_submenu_total++] = "Yes";
					else 
						game_submenu_item[game_submenu_total++] = "No"; 
					game_menu_item[game_menu_total++] = "Flip X"; 
					if (saveOptions.flipx) 
						game_submenu_item[game_submenu_total++] = "Yes";
					else 
						game_submenu_item[game_submenu_total++] = "No"; 
					game_menu_item[game_menu_total++] = "Flip Y"; 
					if (saveOptions.flipy) 
						game_submenu_item[game_submenu_total++] = "Yes";
					else 
						game_submenu_item[game_submenu_total++] = "No"; 
					
					game_menu_item[game_menu_total++] = "Sound";
					if (saveOptions.samplerate) 
						game_submenu_item[game_submenu_total++] = "Yes";
					else 
						game_submenu_item[game_submenu_total++] = "No"; 

					game_menu_item[game_menu_total++] = "Samples";
					if (saveOptions.use_samples) 
						game_submenu_item[game_submenu_total++] = "Yes";
					else 
						game_submenu_item[game_submenu_total++] = "No"; 

					game_menu_item[game_menu_total++] = "Fake Sound";
					if (!save_play_sound) 
						game_submenu_item[game_submenu_total++] = "Yes";
					else 
						game_submenu_item[game_submenu_total++] = "No"; 
						
					game_menu_item[game_menu_total++] = "Artwork";
					if (saveOptions.use_artwork) 
						game_submenu_item[game_submenu_total++] = "Yes";
					else 
						game_submenu_item[game_submenu_total++] = "No"; 

					game_menu_item[game_menu_total++] = "Bitmap Depth";
					if (saveOptions.color_depth == 8) 
						game_submenu_item[game_submenu_total++] = "8-bit";
					else 
						game_submenu_item[game_submenu_total++] = "16-bit"; 

					game_menu_item[game_menu_total++] = "CPU Slice Speedup";
					if (cpu_slices_per_frame_override) 
						game_submenu_item[game_submenu_total++] = "Yes";
					else 
						game_submenu_item[game_submenu_total++] = "No"; 

					game_menu_item[game_menu_total++] = "Auto rotate";
					if (auto_rotate == 0) 
						game_submenu_item[game_submenu_total++] = "Off";
					else if (auto_rotate == 1) 
						game_submenu_item[game_submenu_total++] = "Auto Right"; 
					else if (auto_rotate == 2) 
						game_submenu_item[game_submenu_total++] = "Auto Left"; 

					game_menu_item[game_menu_total++] = "Dirty Rects";
					if (saveUse_dirty == 0) 
						game_submenu_item[game_submenu_total++] = "Off";
					else if (saveUse_dirty == 1) 
						game_submenu_item[game_submenu_total++] = "On"; 
					else if (saveUse_dirty == -1) 
						game_submenu_item[game_submenu_total++] = "Auto"; 

					game_menu_item[game_menu_total++] = "Hide Clones";
					if (hide_clones == 0) 
						game_submenu_item[game_submenu_total++] = "Off";
					else if (hide_clones == 1) 
						game_submenu_item[game_submenu_total++] = "On"; 

					game_menu_item[game_menu_total] = 0;
					game_submenu_item[game_submenu_total] = 0;
					
					ui_displaymenu(Machine->scrbitmap, (const char **)game_menu_item,(const char **)game_submenu_item,0,gamesel,0);
					update_screen(Machine->scrbitmap);

					if (input_ui_pressed_repeat(IPT_UI_DOWN,kREPEAT_TIME))
						gamesel = gamesel + 1;

					if (input_ui_pressed_repeat(IPT_UI_UP,kREPEAT_TIME))
						gamesel = gamesel - 1;
						
					if (gamesel > game_menu_total - 1)
					{
						gamesel = 0;
					}
					else if (gamesel < 0)
					{
						gamesel = game_menu_total - 1;
					}

					if (input_ui_pressed(IPT_UI_SELECT) || input_ui_pressed(IPT_START1))
					{
						switch (gamesel)
						{
							case 0:
							{
								if (savePadOrientation == 0)
									savePadOrientation = 1;
								else if (savePadOrientation == 1)
									savePadOrientation = 2;
								else if (savePadOrientation == 2)
									savePadOrientation = 3;
								else if (savePadOrientation == 3)
									savePadOrientation = 0;
								else
									savePadOrientation = 0;
								break;
							}
							case 1:
							{
								if (saveResolution == -1)
									saveResolution = 0;
								else if (saveResolution == 0)
									saveResolution = 1;
								else if (saveResolution == 1)
									saveResolution = 2;
								else if (saveResolution == 2)
									saveResolution = 3;
								else if (saveResolution == 3)
									saveResolution = -1;
								break;
							}
							case 2:
							{
								if (saveOptions.norotate)
									saveOptions.norotate = 0;
								else
									saveOptions.norotate = 1;
								break;
							}
							case 3:
							{
								if (saveOptions.ror)
									saveOptions.ror = 0;
								else
									saveOptions.ror = 1;
								break;
							}
							case 4:
							{
								if (saveOptions.rol)
									saveOptions.rol = 0;
								else
									saveOptions.rol = 1;
								break;
							}
							case 5:
							{
								if (saveOptions.flipx)
									saveOptions.flipx = 0;
								else
									saveOptions.flipx = 1;
								break;
							}
							case 6:
							{
								if (saveOptions.flipy)
									saveOptions.flipy = 0;
								else
									saveOptions.flipy = 1;
								break;
							}
							case 7:
							{
								if (saveOptions.samplerate)
									saveOptions.samplerate = 0;
								else
									saveOptions.samplerate = 44100;
								break;
							}
							case 8:
							{
								if (saveOptions.use_samples)
									saveOptions.use_samples = 0;
								else
									saveOptions.use_samples = 1;
								break;
							}
							case 9:
							{
								
									save_play_sound = 0;
							}
							case 10:
							{
								if (saveOptions.use_artwork)
									saveOptions.use_artwork = 0;
								else
									saveOptions.use_artwork = 1;
								break;
							}
							
							case 11:
							{
								if (saveOptions.color_depth == 8)
									saveOptions.color_depth = 16;
								else
									saveOptions.color_depth = 8;
								break;
							}
							
							case 12:
							{
								if (cpu_slices_per_frame_override)
									cpu_slices_per_frame_override = 0;
								else
									cpu_slices_per_frame_override = 1;
								break;
							}
							
							case 13:
							{
								if (auto_rotate == 0)
									auto_rotate = 1;
								else if (auto_rotate == 1)
									auto_rotate = 2;
								else if (auto_rotate == 2)
									auto_rotate = 0;
								break;
							}
							
							case 14:
							{
								if (saveUse_dirty == 0)
									saveUse_dirty = 1;
								else if (saveUse_dirty == 1)
									saveUse_dirty = -1;
								else if (saveUse_dirty == -1)
									saveUse_dirty = 0;
								break;
							}
							
							case 15:
							{
								if (hide_clones == 0)
									hide_clones = 1;
								else if (hide_clones == 1)
									hide_clones = 0;
								break;
							}
							
							default:
							{
								break;
							}
						}
						osd_clearbitmap(Machine->scrbitmap);
					}
					
					if (input_ui_pressed(IPT_UI_CANCEL) 
						|| input_ui_pressed(IPT_UI_CONFIGURE))
					{
						osd_clearbitmap(Machine->scrbitmap);
						break;
					}
				}
			}
			
			if (input_ui_pressed_repeat(IPT_UI_DOWN,kREPEAT_TIME))
				sel = sel + 1;

			if (input_ui_pressed_repeat(IPT_UI_UP,kREPEAT_TIME))
				sel = sel - 1;
				
			if (input_ui_pressed_repeat(IPT_UI_RIGHT,kREPEAT_TIME))
				sel = sel + visible - 1;

			if (input_ui_pressed_repeat(IPT_UI_LEFT,kREPEAT_TIME))
				sel = sel - visible + 1;

			if (sel > menu_total - 1)
			{
				sel = menu_total - 1;
			}
			else if (sel < 0)
			{
				sel = 0;
			}
			
			if (input_ui_pressed(IPT_UI_SELECT) || input_ui_pressed(IPT_START1))
			{
				break;
			}
		}

		// find the driver in the list
		for (i = 0;drivers[i];i++)
		{
			if (drivers[i])
			{
				if (drivers[i]->name == menu_item[sel])
				{
					break;
				}
			}
		}

		if (auto_rotate != 0)
		{
			if (drivers[i]->flags & ORIENTATION_SWAP_XY)
			{
				if (auto_rotate == 1)
				{
					savePadOrientation = 1;
					saveOptions.ror = 1;
					saveOptions.norotate = 0;
					saveOptions.rol = 0;
				}
				else if (auto_rotate == 2)
				{
					savePadOrientation = 3;
					saveOptions.ror = 0;
					saveOptions.norotate = 0;
					saveOptions.rol = 1;
				}
			}
			else
			{
				savePadOrientation = 0;
				saveOptions.ror = 0;
				saveOptions.norotate = 0;
				saveOptions.rol = 0;
			}
		}
		
		// add roms until user selects blank item from menu
		saveOptions.image_count = 0;
		for (index = 0; index < MAX_IMAGES; index++)
		{
			saveOptions.image_files[index].name = 0;
			saveOptions.image_files[index].type = IO_END;
		}
		saveOptions.image_files[saveOptions.image_count].type = Validate_ROM("/rd/software/coleco/dkong.rom", drivers[i]->dev);
		saveOptions.image_files[saveOptions.image_count].name = "/rd/software/coleco/dkong";
//		saveOptions.image_count++;
#if 0		
		while(saveOptions.image_count < MAX_IMAGES)
		{
			char directory[255];
			char * ext;
				
			strcpy(directory, "/rd/SOFTWARE/");
			strcat(directory, drivers[i]->name);
			
			saveOptions.image_files[saveOptions.image_count].type 
				= Select_ROM(directory, 
					drivers[i]->dev,
					rom_files[saveOptions.image_count]);
			saveOptions.image_files[saveOptions.image_count].name 
				= rom_files[saveOptions.image_count];
				
			/* cut off extension */
			ext = strrchr(saveOptions.image_files[saveOptions.image_count].name , '.');
			if (ext)
			{
				*ext = 0;
			}
			if (stricmp(saveOptions.image_files[saveOptions.image_count].name,"") != 0)
			{
				saveOptions.image_count++;
			}
			else
			{
				saveOptions.image_files[saveOptions.image_count].name = 0;
				saveOptions.image_files[saveOptions.image_count].type = 0;
				break;
			}
		}
#endif
		// clear the screen
		osd_clearbitmap(Machine->scrbitmap);
		update_screen(Machine->scrbitmap);
		update_screen(Machine->scrbitmap); /* clear both buffers in case we use interlaced mode */

		osd_close_display();
		Machine->scrbitmap = 0;
		
		bitmap_free(Machine->scrbitmap);
		Machine->scrbitmap = 0;
		
		freegfx(Machine->uifont);
		Machine->uifont = 0;

		Machine->gamedrv = 0;
		Machine->drv = 0;
		
		// restore current selected option settings
		gResolution = saveResolution;
		gPadOrientation = savePadOrientation;
		options = saveOptions;
		play_sound = save_play_sound;
		use_dirty = saveUse_dirty; 

		code_close();

		if (drivers[i])
		{
			if (Flight_Check(i) == 0)
			{
				// start the game
				mame( (char *)drivers[i]->name );
			}
		}
		
		// so we refresh the screen shot when we come back
		lastone = -9999;
		hide_menu = 0;
		
		for (i = 0; i < 4; i++)
			osd_led_w(i, 0);
			
		// restore
		use_dirty = saveUse_dirty;
	}
	
	free(menu_item);
}

#else

void mamefront(void)
{
	int sel;
	int visible;
	int hide_menu = 0;
	int lastone = -9999;
	int menu_total;
	char **menu_item;
	int i,j;

	memset(&options, 0, sizeof(options));
	// Init default options
	options.norotate     = 0;
	options.ror          = 0;
	options.rol          = 0;
	options.flipx        = 0;
	options.flipy        = 0;
	options.use_artwork  = 1;
	options.color_depth  = 16;
	options.use_emulated_ym3812 = 1;
	
	options.use_samples  = 1;
	options.samplerate   = 44100;
	play_sound = 1;
	
	use_dirty = 0; // off

	gResolution = 0;
	gPadOrientation = 0;
	cpu_slices_per_frame_override = 1;

    set_config_file ("/CD/MAME.CFG");

	parse_cmdline(0);

	Excess_And_Missing_Check();

	if (save_argc == 2)
	{
		// start the game
		mame( save_argv[1] );
	}
	else
	{
	// first is default selected
	sel = 0;

	while(TRUE)
	{
		struct GameOptions saveOptions;
		int save_play_sound;
		int saveResolution;
		int savePadOrientation;
		struct MachineDriver fake_driver;
		struct GameDriver fake_gamedrv;
		struct InputPort fake_input_port;
		
		{		
			// count drivers so we can allocate enough memory
			menu_total = 0;
			for (i = 0;drivers[i];i++)
			{
				if (drivers[i])
				{
					if (!(hide_clones && ((drivers[i]->clone_of == 0) || (drivers[i]->clone_of != &driver_0))))
					{
						// check if rom exists and add it to list
						if (osd_faccess(drivers[i]->name, OSD_FILETYPE_ROM))
						{
							menu_total++;
						}
					}
				}
			}
			// add one for zero at end of list and add another two for rescan & exit
			menu_item = (char **)malloc(sizeof(char **) * (menu_total + 3));

			// reset count to start again and this time add the ones we find
			menu_total = 0;

			// add one for rescan of CD
			menu_item[menu_total++] = "RESCAN";
			// add one for exit
			menu_item[menu_total++] = "EXIT";

			for (i = 0;drivers[i];i++)
			{
				if (drivers[i])
				{
					if (!(hide_clones && ((drivers[i]->clone_of == 0) || (drivers[i]->clone_of != &driver_0))))
					{
						// check if rom exists and add it to list
						if (osd_faccess(drivers[i]->name, OSD_FILETYPE_ROM))
						{
							if (Flight_Check(i))
							{
								// mark file as not running by making it upper case
								int x;
								char * name;
								
								for (x = 0; x < strlen(drivers[i]->name); x++)
								{
									name = (char *)drivers[i]->name;
									name[x] = toupper(name[x]);
								}
							}
							menu_item[menu_total++] = (char *)drivers[i]->name;
						}
					}
				}
			}
			
			// end of the list
			menu_item[menu_total] = 0;

		//	qsort(menu_item, menu_total, sizeof(char **), sort_func);
		}
		// reset screen to normal orientation and resolution while in main menu
		// we will restore it just before we start the game		

		memset(&fake_driver, 0, sizeof(fake_driver));
		memset(&fake_gamedrv, 0, sizeof(fake_gamedrv));
		memset(&fake_input_port, 0, sizeof(fake_input_port));
		
		saveOptions = options;
		save_play_sound = play_sound;
		saveResolution = gResolution;
		savePadOrientation = gPadOrientation;
		saveUse_dirty = use_dirty; 
		
		options.norotate    = 0;
		options.ror         = 0;
		options.rol         = 0;
		options.flipx       = 0;
		options.flipy       = 0;
		options.samplerate  = 0;
		options.use_samples = 0;
		options.use_artwork = 0;
		options.color_depth  = 16;

//		play_sound = 1;
		gResolution = 0;
		gPadOrientation = 0;

		// Init enough of MAME's structures to get the user menus to work		
		
		Machine->drv = &fake_driver;
		fake_driver.frames_per_second = 60;
		fake_driver.video_attributes = 0; // VIDEO_MODIFIES_PALETTE;
		fake_driver.total_colors = NUMBER_OF_MENU_COLORS;
		fake_driver.screen_width = 320;
		fake_driver.screen_height = 240;

		Machine->gamedrv = &fake_gamedrv;
		
		Machine->orientation = ROT0;
		Machine->ui_orientation = ROT0;
		Machine->color_depth = 16;
		
		Machine->input_ports_default = &fake_input_port;
		Machine->input_ports = &fake_input_port;
		Machine->debug_bitmap = 0;
		fake_input_port.type = IPT_END;
		
		code_init();
		
		Machine->scrbitmap = bitmap_alloc_depth(fake_driver.screen_width,fake_driver.screen_height,Machine->color_depth);
		osd_create_display(fake_driver.screen_width,
			fake_driver.screen_height,
			Machine->color_depth,
			fake_driver.frames_per_second,0,0);
		set_visible_area(0,fake_driver.screen_width,0,fake_driver.screen_height);		
		Machine->uifont = builduifont();
		
		visible = (Machine->uiheight / (3 * Machine->uifontheight / 2))/2 - 1;

		osd_allocate_colors(Machine->drv->total_colors, (void *)dummy_palette, pens, 0, 0, 0);

		while (TRUE)
		{
			if (menu_total == 0) // somethings wrong, nothing in the list, let the user reconfigure and then reboot
			{
			}
#if 1
			if (input_ui_pressed(IPT_COIN1))
			{
				hide_menu = !hide_menu;
				lastone = -9999;
			}
			
			// draw snapshot before menu
			if (sel != lastone)
			{
				int orig_width;
				int orig_height;

				lastone = sel;

				// save two colors for new UI B&W
				snap_load_size_common(menu_item[sel], 0, 32768, Machine->scrbitmap->width, Machine->scrbitmap->height, &snap_overlay, &orig_width, &orig_height);
				if (snap_overlay == NULL)
				{
					// find the driver in the list
					for (i = 0;drivers[i];i++)
					{
						if (drivers[i])
						{
							if (drivers[i]->name == menu_item[sel])
							{
								break;
							}
						}
					}
					
					if (drivers[i])
						if (drivers[i]->clone_of != 0)
							if (drivers[i]->clone_of != &driver_0)
								snap_load_size_common(drivers[i]->clone_of->name, 0, 32768, Machine->scrbitmap->width, Machine->scrbitmap->height, &snap_overlay, &orig_width, &orig_height);
					
					// if it's still NULL
					if (snap_overlay == NULL)
						snap_load_size_common("MAMED", 0, 32768, Machine->scrbitmap->width, Machine->scrbitmap->height, &snap_overlay, &orig_width, &orig_height);
				}

				osd_clearbitmap(Machine->scrbitmap);
				if (snap_overlay)
				{
					int offsetx;
					int offsety;
					
					backdrop_refresh(snap_overlay);
					
//					for (i = 0; i < snap_overlay->num_pens_used; i++)
//						osd_modify_pen(i, snap_overlay->orig_palette[i * 3], 
//							snap_overlay->orig_palette[i * 3 + 1], 
//							snap_overlay->orig_palette[i * 3 + 2]);

					offsetx = (Machine->scrbitmap->width - orig_width) / 2;
					offsety = (Machine->scrbitmap->height - orig_height) / 2;
					
					copybitmap(Machine->scrbitmap, snap_overlay->orig_artwork, 0, 0, offsetx, offsety, 0, TRANSPARENCY_NONE, 0);
					osd_mark_dirty(0,0,Machine->scrbitmap->width, Machine->scrbitmap->height);

					Machine->uifont->colortable[0] = 0;
					Machine->uifont->colortable[1] = ((255 >> 3) << 11) | ((255 >> 2) << 5) | ((255 >> 3) << 0); //snap_overlay->num_pens_used + 1;
					Machine->uifont->colortable[2] = ((255 >> 3) << 11) | ((255 >> 2) << 5) | ((255 >> 3) << 0); //snap_overlay->num_pens_used + 1;
					Machine->uifont->colortable[3] = 0;	

					artwork_free(&snap_overlay);
					snap_overlay = NULL;
				}
				else
				{
					hide_menu = 0;
				}
			}

#endif			
			// draw menu
			if (!hide_menu)
			{		
				ui_displaymenu(Machine->scrbitmap, (const char **)menu_item,0,0,sel,0);
			}

			// get it all to the screen
			update_screen(Machine->scrbitmap);

			if (input_ui_pressed(IPT_UI_CONFIGURE))
			{
				// my little options menu
				const char *game_submenu_item[20];
				int game_submenu_total;
				const char *game_menu_item[20];
				int game_menu_total;
				int gamesel;
				char memStr1[50];
				char memStr2[50];

				// so we refresh the screen shot when we come back
				lastone = -9999;
				osd_clearbitmap(Machine->scrbitmap);

				gamesel = 0;
				
				while (TRUE)
				{
					game_menu_total = 0;
					game_submenu_total = 0;
					
					game_menu_item[game_menu_total++] = "Pad Orientation"; 
					if (savePadOrientation == 0) 
						game_submenu_item[game_submenu_total++] = "Up";
					else if (savePadOrientation == 1)
						game_submenu_item[game_submenu_total++] = "Right";
					else if (savePadOrientation == 2)
						game_submenu_item[game_submenu_total++] = "Down";
					else if (savePadOrientation == 3)
						game_submenu_item[game_submenu_total++] = "Left";
					else
						game_submenu_item[game_submenu_total++] = "Unknown";

					game_menu_item[game_menu_total++] = "Resolution";
					if (saveResolution == 1) 
						game_submenu_item[game_submenu_total++] = "640x480TAFill";
					else if (saveResolution == -1) 
						game_submenu_item[game_submenu_total++] = "640x480TAScale";
					else if (saveResolution == 0) 
						game_submenu_item[game_submenu_total++] = "320x240Direct";
					else if (saveResolution == 2) 
						game_submenu_item[game_submenu_total++] = "320x240TAFill";
					else if (saveResolution == 3) 
						game_submenu_item[game_submenu_total++] = "640x480Direct";
					
					game_menu_item[game_menu_total++] = "No Rotation";
					if (saveOptions.norotate) 
						game_submenu_item[game_submenu_total++] = "Yes";
					else 
						game_submenu_item[game_submenu_total++] = "No"; 
					game_menu_item[game_menu_total++] = "Rotation Right"; 
					if (saveOptions.ror) 
						game_submenu_item[game_submenu_total++] = "Yes";
					else 
						game_submenu_item[game_submenu_total++] = "No"; 
					game_menu_item[game_menu_total++] = "Rotation Left"; 
					if (saveOptions.rol) 
						game_submenu_item[game_submenu_total++] = "Yes";
					else 
						game_submenu_item[game_submenu_total++] = "No"; 
					game_menu_item[game_menu_total++] = "Flip X"; 
					if (saveOptions.flipx) 
						game_submenu_item[game_submenu_total++] = "Yes";
					else 
						game_submenu_item[game_submenu_total++] = "No"; 
					game_menu_item[game_menu_total++] = "Flip Y"; 
					if (saveOptions.flipy) 
						game_submenu_item[game_submenu_total++] = "Yes";
					else 
						game_submenu_item[game_submenu_total++] = "No"; 
					
					game_menu_item[game_menu_total++] = "Sound";
					if (saveOptions.samplerate) 
						game_submenu_item[game_submenu_total++] = "Yes";
					else 
						game_submenu_item[game_submenu_total++] = "No"; 

					game_menu_item[game_menu_total++] = "Samples";
					if (saveOptions.use_samples) 
						game_submenu_item[game_submenu_total++] = "Yes";
					else 
						game_submenu_item[game_submenu_total++] = "No"; 

					game_menu_item[game_menu_total++] = "Fake Sound";
					if (!save_play_sound) 
						game_submenu_item[game_submenu_total++] = "Yes";
					else 
						game_submenu_item[game_submenu_total++] = "No"; 
						
					game_menu_item[game_menu_total++] = "Artwork";
					if (saveOptions.use_artwork) 
						game_submenu_item[game_submenu_total++] = "Yes";
					else 
						game_submenu_item[game_submenu_total++] = "No"; 

					game_menu_item[game_menu_total++] = "Bitmap Depth";
					if (saveOptions.color_depth == 8) 
						game_submenu_item[game_submenu_total++] = "8-bit";
					else 
						game_submenu_item[game_submenu_total++] = "16-bit"; 

					game_menu_item[game_menu_total++] = "CPU Slice Speedup";
					if (cpu_slices_per_frame_override) 
						game_submenu_item[game_submenu_total++] = "Yes";
					else 
						game_submenu_item[game_submenu_total++] = "No"; 

					game_menu_item[game_menu_total++] = "Auto rotate";
					if (auto_rotate == 0) 
						game_submenu_item[game_submenu_total++] = "Off";
					else if (auto_rotate == 1) 
						game_submenu_item[game_submenu_total++] = "Auto Right"; 
					else if (auto_rotate == 2) 
						game_submenu_item[game_submenu_total++] = "Auto Left"; 

					game_menu_item[game_menu_total++] = "Dirty Rects";
					if (saveUse_dirty == 0) 
						game_submenu_item[game_submenu_total++] = "Off";
					else if (saveUse_dirty == 1) 
						game_submenu_item[game_submenu_total++] = "On"; 
					else if (saveUse_dirty == -1) 
						game_submenu_item[game_submenu_total++] = "Auto"; 
					
					game_menu_item[game_menu_total++] = "Hide Clones";
					if (hide_clones == 0) 
						game_submenu_item[game_submenu_total++] = "Off";
					else if (hide_clones == 1) 
						game_submenu_item[game_submenu_total++] = "On"; 

					sprintf(memStr1, "%dK used", total_memory_in_use / 1024);

					game_submenu_item[game_submenu_total++] = "\0"; 
					game_menu_item[game_menu_total++] = memStr1;

					sprintf(memStr2, "%dK max used", max_memory_in_use / 1024);

					game_submenu_item[game_submenu_total++] = "\0"; 
					game_menu_item[game_menu_total++] = memStr2;

					game_menu_item[game_menu_total] = 0;
					game_submenu_item[game_submenu_total] = 0;
					
					ui_displaymenu(Machine->scrbitmap, (const char **)game_menu_item,(const char **)game_submenu_item,0,gamesel,0);
					update_screen(Machine->scrbitmap);

					if (input_ui_pressed_repeat(IPT_UI_DOWN,kREPEAT_TIME))
						gamesel = gamesel + 1;

					if (input_ui_pressed_repeat(IPT_UI_UP,kREPEAT_TIME))
						gamesel = gamesel - 1;
						
					if (gamesel > game_menu_total - 1)
					{
						gamesel = 0;
					}
					else if (gamesel < 0)
					{
						gamesel = game_menu_total - 1;
					}

					if (input_ui_pressed(IPT_UI_SELECT) || input_ui_pressed(IPT_START1))
					{
						switch (gamesel)
						{
							case 0:
							{
								if (savePadOrientation == 0)
									savePadOrientation = 1;
								else if (savePadOrientation == 1)
									savePadOrientation = 2;
								else if (savePadOrientation == 2)
									savePadOrientation = 3;
								else if (savePadOrientation == 3)
									savePadOrientation = 0;
								else
									savePadOrientation = 0;
								break;
							}
							case 1:
							{
								if (saveResolution == -1)
									saveResolution = 0;
								else if (saveResolution == 0)
									saveResolution = 1;
								else if (saveResolution == 1)
									saveResolution = 2;
								else if (saveResolution == 2)
									saveResolution = 3;
								else if (saveResolution == 3)
									saveResolution = -1;
								break;
							}
							case 2:
							{
								if (saveOptions.norotate)
									saveOptions.norotate = 0;
								else
									saveOptions.norotate = 1;
								break;
							}
							case 3:
							{
								if (saveOptions.ror)
									saveOptions.ror = 0;
								else
									saveOptions.ror = 1;
								break;
							}
							case 4:
							{
								if (saveOptions.rol)
									saveOptions.rol = 0;
								else
									saveOptions.rol = 1;
								break;
							}
							case 5:
							{
								if (saveOptions.flipx)
									saveOptions.flipx = 0;
								else
									saveOptions.flipx = 1;
								break;
							}
							case 6:
							{
								if (saveOptions.flipy)
									saveOptions.flipy = 0;
								else
									saveOptions.flipy = 1;
								break;
							}
							case 7:
							{
								if (saveOptions.samplerate)
									saveOptions.samplerate = 0;
								else
									saveOptions.samplerate = 11025;
								break;
							}
							case 8:
							{
								if (saveOptions.use_samples)
									saveOptions.use_samples = 0;
								else
									saveOptions.use_samples = 1;
								break;
							}
							case 9:
							{
								if (save_play_sound)
									save_play_sound = 0;
								else
									save_play_sound = 1;
								break;
							}
							case 10:
							{
								if (saveOptions.use_artwork)
									saveOptions.use_artwork = 0;
								else
									saveOptions.use_artwork = 1;
								break;
							}
							case 11:
							{
								if (saveOptions.color_depth == 8)
									saveOptions.color_depth = 16;
								else
									saveOptions.color_depth = 8;
								break;
							}
							
							case 12:
							{
								if (cpu_slices_per_frame_override)
									cpu_slices_per_frame_override = 0;
								else
									cpu_slices_per_frame_override = 1;
								break;
							}
							
							case 13:
							{
								if (auto_rotate == 0)
									auto_rotate = 1;
								else if (auto_rotate == 1)
									auto_rotate = 2;
								else if (auto_rotate == 2)
									auto_rotate = 0;
								break;
							}
														
							case 14:
							{
								if (saveUse_dirty == 0)
									saveUse_dirty = 1;
								else if (saveUse_dirty == 1)
									saveUse_dirty = -1;
								else if (saveUse_dirty == -1)
									saveUse_dirty = 0;
								break;
							}
							
							case 15:
							{
								if (hide_clones == 0)
									hide_clones = 1;
								else if (hide_clones == 1)
									hide_clones = 0;
								break;
							}
							
							default:
							{
								break;
							}
						}
						osd_clearbitmap(Machine->scrbitmap);
					}
					
					if (input_ui_pressed(IPT_UI_CANCEL) 
						|| input_ui_pressed(IPT_UI_CONFIGURE))
					{
						osd_clearbitmap(Machine->scrbitmap);
						break;
					}
				}
			}
			
			if (input_ui_pressed_repeat(IPT_UI_DOWN,kREPEAT_TIME))
				sel = sel + 1;

			if (input_ui_pressed_repeat(IPT_UI_UP,kREPEAT_TIME))
				sel = sel - 1;
				
			if (input_ui_pressed_repeat(IPT_UI_RIGHT,kREPEAT_TIME))
				sel = sel + visible - 1;

			if (input_ui_pressed_repeat(IPT_UI_LEFT,kREPEAT_TIME))
				sel = sel - visible + 1;

			if (sel > menu_total - 1)
			{
				sel = 0;
			}
			else if (sel < 0)
			{
				sel = menu_total - 1;
			}

			if (input_ui_pressed(IPT_UI_SELECT) || input_ui_pressed(IPT_START1))
			{
				break;
			}
		}

		// find the driver in the list
		for (i = 0;drivers[i];i++)
		{
			if (drivers[i])
			{
				if (drivers[i]->name == menu_item[sel])
				{
					break;
				}
			}
		}

		if (auto_rotate != 0)
		{
			if (drivers[i])
			{
				if (drivers[i]->flags & ORIENTATION_SWAP_XY)
				{
					if (auto_rotate == 1)
					{
						savePadOrientation = 1;
						saveOptions.ror = 1;
						saveOptions.norotate = 0;
						saveOptions.rol = 0;
					}
					else if (auto_rotate == 2)
					{
						savePadOrientation = 3;
						saveOptions.ror = 0;
						saveOptions.norotate = 0;
						saveOptions.rol = 1;
					}
				}
				else
				{
					savePadOrientation = 0;
					saveOptions.ror = 0;
					saveOptions.norotate = 0;
					saveOptions.rol = 0;
				}
			}
			else
			{
				savePadOrientation = 0;
				saveOptions.ror = 0;
				saveOptions.norotate = 0;
				saveOptions.rol = 0;
			}
		}
		
		// clear the screen
		osd_clearbitmap(Machine->scrbitmap);
		update_screen(Machine->scrbitmap);
		update_screen(Machine->scrbitmap); /* clear both buffers in case we use interlaced mode */
	
		osd_close_display();
		
		bitmap_free(Machine->scrbitmap);
		Machine->scrbitmap = 0;

		freegfx(Machine->uifont);
		Machine->uifont = 0;

		Machine->gamedrv = 0;
		Machine->drv = 0;
		
		// restore current selected option settings
		gResolution = saveResolution;
		gPadOrientation = savePadOrientation;
		options = saveOptions;
		play_sound = save_play_sound;
		use_dirty = saveUse_dirty;
		
		code_close();

		if ( drivers[i] )
		{
			if (Flight_Check(i) == 0)
			{
				// start the game
				mame( (char *)drivers[i]->name );
			}
			else
			{
				printf("Failed flight check\n");
			}
		}
		else
		{
			if (stricmp( menu_item[sel], "RESCAN") == 0)
			{
				uint32 hnd;
				
				printf("No rom selected\n");

				/* Force a rescan of the CD, boot CD must have a roms directory for this to work */
				hnd = fs->open("/cd/", O_RDONLY | O_DIR);
				if (hnd) 
				{
					fs->ioctl(hnd, NULL, 0);
					fs->close(hnd);
				}
				
			    hnd = fs->open("/cd", O_RDONLY | O_DIR);
			    fs->close(hnd);

				// first is default selected
				sel = 0;
			}
			else if (stricmp( menu_item[sel], "EXIT") == 0)
			{
				free(menu_item);
				return;
			}
		}
		
		// so we refresh the screen shot when we come back
		lastone = -9999;
		hide_menu = 0;
		
		for (i = 0; i < 4; i++)
			osd_led_w(i, 0);
			
		// restore
		use_dirty = saveUse_dirty;

		// we are going to recreate it in a second
		free(menu_item);
	}
	}
}
#endif

/* not a standard ANSI function but MAME uses it so stick it here */
int stricmp(const char *inString1, const char *inString2)
{
    char c1, c2;
    while (1)
    {
    	if (isupper(*inString1))
			c1 = tolower(*inString1);
		else
			c1 = *inString1;
		inString1++;
    	if (isupper(*inString2))
    		c2 = tolower(*inString2);
    	else
    		c2 = *inString2;
    	inString2++;
        if (c1 < c2) return -1;
        if (c1 > c2) return 1;
        if (c1 == 0) return 0;
    }
}

/* not a standard ANSI function but MAME uses it so stick it here */
int strnicmp(const char *inString1, const char *inString2, int len)
{
    char c1, c2;
    int i;
    
    for (i = 0; i < len; i++)
    {
    	if (isupper(*inString1))
			c1 = tolower(*inString1);
		else
			c1 = *inString1;
		inString1++;
    	if (isupper(*inString2))
    		c2 = tolower(*inString2);
    	else
    		c2 = *inString2;
    	inString2++;
        if (c1 < c2) return -1;
        if (c1 > c2) return 1;
        if (c1 == 0) return 0;
    }
    
    return 0;
}

void DIE_OR_NOT(void)
{
//	while (1)
//		;
}

static void init_memory_table(void)
{
	int i;
	
	for (i = 0; i < MAX_MEMORY_TABLE; i++)
	{
		memory_table[i].address = 0;
		memory_table[i].size = 0;
		memory_table[i].in_use = 0;
		memory_table[i].count = 0;
	}
	
	current_malloc_count = 0;
	max_memory_in_use = 0;
	total_memory_in_use = 0;
	
	// point the pointer to us
	oldmalloc = fabi->malloc;
	oldfree = fabi->free;
	oldmemalign = fabi->memalign;
	fabi->malloc = mamed_malloc;
	fabi->free = (void *)mamed_free;
	fabi->memalign = mamed_memalign;
	
	printf("memory table initialized\n");
}

void check_memory_table(void)
{
#if 0	
	int i;
	
	for (i = 0; i < MAX_MEMORY_TABLE; i++)
	{
		if (memory_table[i].in_use == 1)
		{
			// check head
			if (*(unsigned long *)(memory_table[i].address) != HEAD_FLAG)
			{
				printf("check memory, head doesn't match\n");
				DIE_OR_NOT();
			}

			if (*(unsigned long *)(memory_table[i].address + memory_table[i].size - 4) != TAIL_FLAG)
			{
				printf("check memory, tail doesn't match\n");
				DIE_OR_NOT();
			}
		}
	}
#endif
}

static void shutdown_memory_table(void)
{
	int i;
	int count = 0, total = 0;
	
	// put the pointer back
	fabi->malloc = oldmalloc;
	fabi->free = oldfree;
	fabi->memalign = oldmemalign;
	
	printf("memory table shutdown\n");
	printf("max %d; current %d\n", max_memory_in_use, total_memory_in_use);
	for (i = 0; i < MAX_MEMORY_TABLE; i++)
	{
		if (memory_table[i].in_use == 1)
		{
			count++;
			total = total + memory_table[i].size;
		}
	}
	printf("blocks in use %d, total memory %d, number of allocs %d\n", count, total, current_malloc_count);

	for (i = 0; i < MAX_MEMORY_TABLE; i++)
	{
		if (memory_table[i].in_use == 1)
		{
			printf("Addess %d, size %d, count %d %s\n", memory_table[i].address, memory_table[i].size, memory_table[i].count, memory_table[i].address + 4);
		}
	}

	check_memory_table();
}

static void* mamed_malloc(size_t size)
{
	int i;
	char * p;

	check_memory_table();
	
	// check for zero length
	if (size == 0)
	{
		printf("malloc(0)\n");
		return 0;
	}
	
	// adjust size for marking head & tail on a long word boundry
	size = 4 - (size % 4) + size + 4 + 4;
	
	p = oldmalloc(size);
	
	if (p == 0)
	{
		printf("malloc(%d) = 0\n", size);
		return 0;
	}
	
	// clear memory just to be safe
//	memset(p, 0xff, size);
	
	// mark head and tail
	*(unsigned long *)(p) = HEAD_FLAG;
	*(unsigned long *)(p + size - 4) = TAIL_FLAG;
	
//	printf("%d = malloc(%d)\n", p, size);

	// check for overlap
#if 0
	for (i = 0; i < MAX_MEMORY_TABLE; i++)
	{
		if (memory_table[i].in_use)
		{
			if ((memory_table[i].address < p) && ((memory_table[i].address + memory_table[i].size) > p))
			{
				printf("overlap occured, front overlaps back of existing block\n");
				DIE_OR_NOT();
			}
			if ((p < memory_table[i].address) && ((p + size) > memory_table[i].address))
			{
				printf("overlap occured, back overlaps front of existing block \n");
				DIE_OR_NOT();
			}
			if ((p < memory_table[i].address) && ((p + size) > (memory_table[i].address + memory_table[i].size)))
			{
				printf("overlap occured, new block overlaps existing block completely\n");
				DIE_OR_NOT();
			}
		}
	}
#endif
	
	// find and empty spot
	for (i = 0; i < MAX_MEMORY_TABLE; i++)
	{
		if (memory_table[i].in_use == 0)
		{
			break;
		}
	}
	
	// ran out of room in table
	if (i == MAX_MEMORY_TABLE)
	{
		printf("ran out of room in table\n");
		DIE_OR_NOT();
	}
	
	// store result
	memory_table[i].address = p;
	memory_table[i].size = size;
	memory_table[i].in_use = 1;
	memory_table[i].count = current_malloc_count++;

	// keep total up to date
	total_memory_in_use = total_memory_in_use + size;

	// check for ceiling
	if (total_memory_in_use > max_memory_in_use)
	{
		max_memory_in_use = total_memory_in_use;
	}

	return p + 4;
}

static void mamed_free(char *p)
{
	int i;
	
	check_memory_table();

	// check for bad pointer
	if (p == 0)
	{
//		printf("free(0)\n");
		return;
	}

	// adjust pointer
	p = p - 4;
	
//	printf("free(%d)\n", p);

	// find it in the table
	for (i = 0; i < MAX_MEMORY_TABLE; i++)
	{
		if (memory_table[i].address == p)
		{
			break;
		}
	}

	// could not find it in table
	if (i == MAX_MEMORY_TABLE)
	{
		printf("could not find it in table\n");
		DIE_OR_NOT();
		return;
	}
	
	// check head
	if (*(unsigned long *)(p) != HEAD_FLAG)
	{
		printf("free memory, head doesn't match\n");
		DIE_OR_NOT();
	}

	// check tail
	if (*(unsigned long *)(p + memory_table[i].size - 4) != TAIL_FLAG)
	{
		printf("free memory, tail doesn't match\n");
		DIE_OR_NOT();
	}

	// clear it to be sure no one uses it again
//	memset(memory_table[i].address, 0, memory_table[i].size);
	
	// keep total up to date
	total_memory_in_use = total_memory_in_use - memory_table[i].size;
	
	// store result
	memory_table[i].address = 0;
	memory_table[i].size = 0;
	memory_table[i].in_use = 0;
	memory_table[i].count = 0;

	// really free the memory
	oldfree(p);
}

static void* mamed_memalign(size_t size, size_t align)
{
	int i;
	void * p;

	check_memory_table();
	
	// check for zero length
	if (size == 0)
	{
//		printf("malloc(0)\n");
		return 0;
	}
	
	// adjust size for marking head & tail on a long word boundry
	size = 4 - (size % 4) + size + 4 + 4;

	p = oldmemalign(size,align);

	if (p == 0)
	{
		printf("malloc(%d) = 0\n", size);
		return 0;
	}
	
	// clear memory just to be safe
//	memset(p, 0xff, size);
	
	// mark head and tail
	*(unsigned long *)(p) = HEAD_FLAG;
	*(unsigned long *)(p + size - 4) = TAIL_FLAG;
	
//	printf("mamed_memalign(%d, %d)\n", size, align);

#if 0
	// check for overlap
	for (i = 0; i < MAX_MEMORY_TABLE; i++)
	{
		if (memory_table[i].in_use)
		{
			if ((memory_table[i].address < p) && ((memory_table[i].address + memory_table[i].size) > p))
			{
				printf("overlap occured, front overlaps back of existing block\n");
				DIE_OR_NOT();
			}
			if ((p < memory_table[i].address) && ((p + size) > memory_table[i].address))
			{
				printf("overlap occured, back overlaps front of existing block \n");
				DIE_OR_NOT();
			}
			if ((p < memory_table[i].address) && ((p + size) > (memory_table[i].address + memory_table[i].size)))
			{
				printf("overlap occured, new block overlaps existing block completely\n");
				DIE_OR_NOT();
			}
		}
	}
#endif	
	// find and empty spot
	for (i = 0; i < MAX_MEMORY_TABLE; i++)
	{
		if (memory_table[i].in_use == 0)
		{
			break;
		}
	}
	
	// ran out of room in table
	if (i == MAX_MEMORY_TABLE)
	{
		printf("ran out of room in table\n");
		DIE_OR_NOT();
	}
	
	// store result
	memory_table[i].address = p;
	memory_table[i].size = size;
	memory_table[i].in_use = 1;
	memory_table[i].count = current_malloc_count++;

	// check for ceiling
	total_memory_in_use = total_memory_in_use + size;
	if (total_memory_in_use > max_memory_in_use)
	{
		max_memory_in_use = total_memory_in_use;
	}
	
	return p;
}
