#include "driver.h"
#include "mamedbg.h"
#include "vidhrdw/vector.h"
#include "dirty.h"
#include "ticker.h"


#include "osdepend.h"

static int warming_up;

/* tweak values for centering tweaked modes */
int center_x;
int center_y;

#define BACKGROUND 0


dirtygrid grid1;
dirtygrid grid2;
char *dirty_old=grid1;
char *dirty_new=grid2;

static void update_screen_dummy(struct osd_bitmap *bitmap);
void (*update_screen)(struct osd_bitmap *bitmap) = update_screen_dummy;
void (*update_screen_debugger)(struct osd_bitmap *bitmap) = update_screen_dummy;
void osd_clearbitmap(struct osd_bitmap *bitmap);


static int video_depth,video_fps,video_attributes,video_orientation;
static int modifiable_palette;
static int screen_colors;
static UINT8 *current_palette;
static const UINT8 *dbg_palette;
static unsigned int *dirtycolor;
static int dirtypalette;
static int dirty_bright;
static int bright_lookup[256];

int frameskip,autoframeskip;
#define FRAMESKIP_LEVELS 12

int vsync_frame_rate;
int skiplines;
int skipcolumns;
int use_dirty;
float osd_gamma_correction = 1.0;
int brightness;
float brightness_paused_adjust;
int gfx_width;
int gfx_height;
static int vis_min_x,vis_max_x,vis_min_y,vis_max_y;


static int viswidth;
static int visheight;
static int skiplinesmax;
static int skipcolumnsmax;
static int skiplinesmin;
static int skipcolumnsmin;
static int show_debugger,debugger_focus_changed;

static int vector_game;


int gfx_xoffset;
int gfx_yoffset;
int gfx_display_lines;
int gfx_display_columns;
static int xmultiply,ymultiply;
int throttle = 1;       /* toggled by F10 */

static int gone_to_gfx_mode;
static int frameskip_counter;
static int frames_displayed;
static TICKER start_time,end_time;    /* to calculate fps average on exit */
#define FRAMES_TO_SKIP 20       /* skip the first few frames from the FPS calculation */
							/* to avoid counting the copyright and info screens */

unsigned short * current_display = NULL;

static UINT16 *current_pens_lookup = 0;
static UINT16 *pens_lookup = 0;
static UINT16 *debug_pens_lookup = 0;
static uint32 mame_txr;

extern int gResolution;

/* Create a bitmap. */
/* VERY IMPORTANT: the function must allocate also a "safety area" 16 pixels wide all */
/* around the bitmap. This is required because, for performance reasons, some graphic */
/* routines don't clip at boundaries of the bitmap. */

const int safety = 16;

// built in memcpy sucks, this is tons faster
void mymemcpy(long * dst, long * src, int num)
{
	int i;
	
	num >>= 2;
	
	for (i = 0; i < num; i++)
	{
		*dst++ = *src++;
	}
}

void update_screen_nodirty_16(struct osd_bitmap *bitmap)
{
	register unsigned short * walk_display;
	register unsigned short * walk_bitmap;
	register int i;
	register int j;
	int startRow;
	int endRow;
	int startColumn;
	int endColumn;
	int displayInc;

	current_display = video->vram_s;

	startRow = skiplines;
	endRow = skiplines + gfx_display_lines;
	startColumn = skipcolumns;
	endColumn = skipcolumns + gfx_display_columns;
	
	walk_display = current_display + gfx_xoffset + (gfx_width * gfx_yoffset);
	displayInc = gfx_width - gfx_display_columns;
	
	walk_bitmap = bitmap->line[startRow] + skipcolumns;
	
	if (current_pens_lookup)
	{
		for (i = startRow;i < endRow;i++)
		{
			for (j = startColumn; j < endColumn; j++)
			{
				*walk_display++ = current_pens_lookup[*walk_bitmap++];
			}
			
			walk_display += displayInc;
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}
	else
	{
		for (i = startRow;i < endRow;i++)
		{
			mymemcpy(walk_display, walk_bitmap, gfx_display_columns * 2);
			walk_display += gfx_display_columns + displayInc;
			
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}
}

void update_screen_dirty_16(struct osd_bitmap *bitmap)
{
	register unsigned short * walk_display;
	register unsigned short * walk_bitmap;
	register int i;
	register int j;
	int startRow;
	int endRow;
	int startColumn;
	int endColumn;
	int displayInc;
	int skipInc;

	current_display = video->vram_s;

	startRow = skiplines;
	endRow = skiplines + gfx_display_lines;
	startColumn = skipcolumns;
	endColumn = skipcolumns + gfx_display_columns;
	
	walk_display = current_display + gfx_xoffset + (gfx_width * gfx_yoffset);
	displayInc = gfx_width - gfx_display_columns;
	skipInc = gfx_width;
	
	walk_bitmap = bitmap->line[startRow] + skipcolumns;
	
	if (current_pens_lookup)
	{
		for (i = startRow;i < endRow;i++)
		{
			if (ISDIRTY(i))
			{
				for (j = startColumn; j < endColumn; j++)
				{
					*walk_display++ = current_pens_lookup[*walk_bitmap++];
				}
				
				walk_display += displayInc;
			}
			else
			{
				walk_display += skipInc;
			}
			
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}
	else
	{
		for (i = startRow;i < endRow;i++)
		{
			if (ISDIRTY(i))
			{
				mymemcpy(walk_display, walk_bitmap, gfx_display_columns * 2);		
			}
			walk_display += skipInc;
			
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}
}

void update_screen_nodirty_ta_320x240_16(struct osd_bitmap *bitmap)
{
	register unsigned short * walk_display;
	register unsigned short * walk_bitmap;
	register int i;
	register int j;
	int startRow;
	int endRow;
	int startColumn;
	int endColumn;
	int displayInc;

	current_display = ta->txr_map(mame_txr);

	startRow = skiplines;
	endRow = skiplines + gfx_display_lines;
	startColumn = skipcolumns;
	endColumn = skipcolumns + gfx_display_columns;
	
	walk_display = current_display; // + (gfx_xoffset/4) + (gfx_width * (gfx_yoffset / 2));
	displayInc = gfx_width - gfx_display_columns;
	
	walk_bitmap = bitmap->line[startRow] + skipcolumns;
	
	if (current_pens_lookup)
	{
		for (i = startRow;i < endRow;i++)
		{
			for (j = startColumn; j < endColumn; j++)
			{
				*walk_display++ = current_pens_lookup[*walk_bitmap++];
			}
			
			walk_display += displayInc;
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}
	else
	{
		for (i = startRow;i < endRow;i++)
		{
			mymemcpy(walk_display, walk_bitmap, gfx_display_columns * 2);
			walk_display += gfx_display_columns + displayInc;
			
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}

  ta->finish_frame();

  ta->begin_render();
  {
  	vertex_ot_t	v;
  	poly_hdr_t	hdr;
    float fade=1.0f;
    float ratio_h=(float)bitmap->height/((240.0f*(float)gfx_height)/240.0f);
    float ratio_w=(float)bitmap->width/((320.0f*(float)gfx_width)/320.0f);

  	ta->poly_hdr_txr(&hdr, TA_OPAQUE, TA_RGB565, gfx_width, gfx_height, mame_txr, /*TA_NO_FILTER*/TA_BILINEAR_FILTER);
  	ta->commit_poly_hdr(&hdr);

  	/* - -
  	   + - */
  	v.flags = TA_VERTEX_NORMAL;
  	v.x = 0.0f;
  	v.y = 240.0f;
  	v.z = 1.0f;
  	v.u = 0.0f; v.v = ratio_h; 
  	v.a = 1.0f; v.r = v.g = v.b = fade;
  	v.oa = v.or = v.og = v.ob = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* + -
  	   - - */
  	v.y = 0.0f;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* - -
  	   - + */
  	v.x = 320.0f;
  	v.y = 240.0f;
  	v.u = ratio_w; v.v = ratio_h; 
  	ta->commit_vertex(&v, sizeof(v));

  	/* - +
  	   - - */
  	v.flags = TA_VERTEX_EOL;
  	v.y = 0.0f;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));
  }
  ta->commit_eol();
  {
    poly_hdr_t	hdr;
    ta->poly_hdr_col(&hdr, TA_TRANSLUCENT);
    ta->commit_poly_hdr(&hdr);
  }
  ta->commit_eol();
}

void update_screen_nodirty_ta_640x480_16(struct osd_bitmap *bitmap)
{
	register unsigned short * walk_display;
	register unsigned short * walk_bitmap;
	register int i;
	register int j;
	int startRow;
	int endRow;
	int startColumn;
	int endColumn;
	int displayInc;

	current_display = ta->txr_map(mame_txr);

	startRow = skiplines;
	endRow = skiplines + gfx_display_lines;
	startColumn = skipcolumns;
	endColumn = skipcolumns + gfx_display_columns;
	
	walk_display = current_display; // + (gfx_xoffset/4) + (gfx_width * (gfx_yoffset / 2));
	displayInc = gfx_width - gfx_display_columns;
	
	walk_bitmap = bitmap->line[startRow] + skipcolumns;
	
	if (current_pens_lookup)
	{
		for (i = startRow;i < endRow;i++)
		{
			for (j = startColumn; j < endColumn; j++)
			{
				*walk_display++ = current_pens_lookup[*walk_bitmap++];
			}
			
			walk_display += displayInc;
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}
	else
	{
		for (i = startRow;i < endRow;i++)
		{
			mymemcpy(walk_display, walk_bitmap, gfx_display_columns * 2);
			walk_display += gfx_display_columns + displayInc;
			
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}

  ta->finish_frame();

  ta->begin_render();
  {
  	vertex_ot_t	v;
  	poly_hdr_t	hdr;
    float fade=1.0f;
    float ratio_h=(float)bitmap->height/((480.0f*(float)gfx_height)/480.0f);
    float ratio_w=(float)bitmap->width/((640.0f*(float)gfx_width)/640.0f);

  	ta->poly_hdr_txr(&hdr, TA_OPAQUE, TA_RGB565, gfx_width, gfx_height, mame_txr, /*TA_NO_FILTER*/TA_BILINEAR_FILTER);
  	ta->commit_poly_hdr(&hdr);

  	/* - -
  	   + - */
  	v.flags = TA_VERTEX_NORMAL;
  	v.x = 0.0f;
  	v.y = 480.0f;
  	v.z = 1.0f;
  	v.u = 0.0f; v.v = ratio_h; 
  	v.a = 1.0f; v.r = v.g = v.b = fade;
  	v.oa = v.or = v.og = v.ob = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* + -
  	   - - */
  	v.y = 0.0f;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* - -
  	   - + */
  	v.x = 640.0f;
  	v.y = 480.0f;
  	v.u = ratio_w; v.v = ratio_h; 
  	ta->commit_vertex(&v, sizeof(v));

  	/* - +
  	   - - */
  	v.flags = TA_VERTEX_EOL;
  	v.y = 0.0f;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));
  }
  ta->commit_eol();
  {
    poly_hdr_t	hdr;
    ta->poly_hdr_col(&hdr, TA_TRANSLUCENT);
    ta->commit_poly_hdr(&hdr);
  }
  ta->commit_eol();
}

void update_screen_nodirty_ta_640x480_16_scaled(struct osd_bitmap *bitmap)
{
	register unsigned short * walk_display;
	register unsigned short * walk_bitmap;
	register int i;
	register int j;
	int startRow;
	int endRow;
	int startColumn;
	int endColumn;
	int displayInc;

	current_display = ta->txr_map(mame_txr);

	startRow = skiplines;
	endRow = skiplines + gfx_display_lines;
	startColumn = skipcolumns;
	endColumn = skipcolumns + gfx_display_columns;
	
	walk_display = current_display; // + (gfx_xoffset/4) + (gfx_width * (gfx_yoffset / 2));
	displayInc = gfx_width - gfx_display_columns;
	
	walk_bitmap = bitmap->line[startRow] + skipcolumns;
	
	if (current_pens_lookup)
	{
		for (i = startRow;i < endRow;i++)
		{
			for (j = startColumn; j < endColumn; j++)
			{
				*walk_display++ = current_pens_lookup[*walk_bitmap++];
			}
			
			walk_display += displayInc;
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}
	else
	{
		for (i = startRow;i < endRow;i++)
		{
			mymemcpy(walk_display, walk_bitmap, gfx_display_columns * 2);
			walk_display += gfx_display_columns + displayInc;
			
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}

  ta->finish_frame();

  ta->begin_render();
  {
  	vertex_ot_t	v;
  	poly_hdr_t	hdr;
    float fade=1.0f;
    float ratio_h;
    float ratio_w;
    float adjust_x;
    float adjust_y;

	if (bitmap->height > bitmap->width)
	{
		ratio_h=(float)bitmap->height/(float)gfx_height;
		ratio_w=ratio_h;//(float)gfx_width/(float)640.0f;
		adjust_x = (gfx_width - bitmap->width)/4;
		adjust_y = 0.0f;
	}
	else
	{
		ratio_w=(float)bitmap->width/(float)gfx_width;
		ratio_h=ratio_w;//(float)gfx_height/(float)480.0f;
		adjust_x = 0.0f;
		adjust_y = (gfx_height - bitmap->height)/4;
	}

  	ta->poly_hdr_txr(&hdr, TA_OPAQUE, TA_RGB565, gfx_width, gfx_height, mame_txr, /*TA_NO_FILTER*/TA_BILINEAR_FILTER);
  	ta->commit_poly_hdr(&hdr);

  	/* - -
  	   + - */
  	v.flags = TA_VERTEX_NORMAL;
  	v.x = 0.0f + adjust_x;
  	v.y = 480.0f + adjust_y;
  	v.z = 1.0f;
  	v.u = 0.0f; v.v = ratio_h; 
  	v.a = 1.0f; v.r = v.g = v.b = fade;
  	v.oa = v.or = v.og = v.ob = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* + -
  	   - - */
  	v.y = 0.0f + adjust_y;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* - -
  	   - + */
  	v.x = 640.0f + adjust_x;
  	v.y = 480.0f + adjust_y;
  	v.u = ratio_w; v.v = ratio_h; 
  	ta->commit_vertex(&v, sizeof(v));

  	/* - +
  	   - - */
  	v.flags = TA_VERTEX_EOL;
  	v.y = 0.0f + adjust_y;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));
  }
  ta->commit_eol();
  {
    poly_hdr_t	hdr;
    ta->poly_hdr_col(&hdr, TA_TRANSLUCENT);
    ta->commit_poly_hdr(&hdr);
  }
  ta->commit_eol();
}

void update_screen_dirty_ta_320x240_16(struct osd_bitmap *bitmap)
{
	register unsigned short * walk_display;
	register unsigned short * walk_bitmap;
	register int i;
	register int j;
	int startRow;
	int endRow;
	int startColumn;
	int endColumn;
	int displayInc;
	int skipInc;

	current_display = ta->txr_map(mame_txr);

	startRow = skiplines;
	endRow = skiplines + gfx_display_lines;
	startColumn = skipcolumns;
	endColumn = skipcolumns + gfx_display_columns;
	
	walk_display = current_display; // + (gfx_xoffset/4) + (gfx_width * (gfx_yoffset / 2));
	displayInc = gfx_width - gfx_display_columns;
	
	walk_bitmap = bitmap->line[startRow] + skipcolumns;
	skipInc = gfx_width;
	
	if (current_pens_lookup)
	{
		for (i = startRow;i < endRow;i++)
		{
			if (ISDIRTY(i))
			{
				for (j = startColumn; j < endColumn; j++)
				{
					*walk_display++ = current_pens_lookup[*walk_bitmap++];
				}
				
				walk_display += displayInc;
			}
			else
			{
				walk_display += skipInc;
			}
			
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}
	else
	{
		for (i = startRow;i < endRow;i++)
		{
			if (ISDIRTY(i))
			{
				mymemcpy(walk_display, walk_bitmap, gfx_display_columns * 2);		
			}
			walk_display += skipInc;
			
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}

  ta->finish_frame();

  ta->begin_render();
  {
  	vertex_ot_t	v;
  	poly_hdr_t	hdr;
    float fade=1.0f;
    float ratio_h=(float)bitmap->height/((240.0f*(float)gfx_height)/240.0f);
    float ratio_w=(float)bitmap->width/((320.0f*(float)gfx_width)/320.0f);

  	ta->poly_hdr_txr(&hdr, TA_OPAQUE, TA_RGB565, gfx_width, gfx_height, mame_txr, /*TA_NO_FILTER*/TA_BILINEAR_FILTER);
  	ta->commit_poly_hdr(&hdr);

  	/* - -
  	   + - */
  	v.flags = TA_VERTEX_NORMAL;
  	v.x = 0.0f;
  	v.y = 240.0f;
  	v.z = 1.0f;
  	v.u = 0.0f; v.v = ratio_h; 
  	v.a = 1.0f; v.r = v.g = v.b = fade;
  	v.oa = v.or = v.og = v.ob = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* + -
  	   - - */
  	v.y = 0.0f;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* - -
  	   - + */
  	v.x = 320.0f;
  	v.y = 240.0f;
  	v.u = ratio_w; v.v = ratio_h; 
  	ta->commit_vertex(&v, sizeof(v));

  	/* - +
  	   - - */
  	v.flags = TA_VERTEX_EOL;
  	v.y = 0.0f;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));
  }
  ta->commit_eol();
  {
    poly_hdr_t	hdr;
    ta->poly_hdr_col(&hdr, TA_TRANSLUCENT);
    ta->commit_poly_hdr(&hdr);
  }
  ta->commit_eol();
}

void update_screen_dirty_ta_640x480_16(struct osd_bitmap *bitmap)
{
	register unsigned short * walk_display;
	register unsigned short * walk_bitmap;
	register int i;
	register int j;
	int startRow;
	int endRow;
	int startColumn;
	int endColumn;
	int displayInc;
	int skipInc;

	current_display = ta->txr_map(mame_txr);

	startRow = skiplines;
	endRow = skiplines + gfx_display_lines;
	startColumn = skipcolumns;
	endColumn = skipcolumns + gfx_display_columns;
	
	walk_display = current_display; // + (gfx_xoffset/4) + (gfx_width * (gfx_yoffset / 2));
	displayInc = gfx_width - gfx_display_columns;
	
	walk_bitmap = bitmap->line[startRow] + skipcolumns;
	skipInc = gfx_width;
	
	if (current_pens_lookup)
	{
		for (i = startRow;i < endRow;i++)
		{
			if (ISDIRTY(i))
			{
				for (j = startColumn; j < endColumn; j++)
				{
					*walk_display++ = current_pens_lookup[*walk_bitmap++];
				}
				
				walk_display += displayInc;
			}
			else
			{
				walk_display += skipInc;
			}
			
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}
	else
	{
		for (i = startRow;i < endRow;i++)
		{
			if (ISDIRTY(i))
			{
				mymemcpy(walk_display, walk_bitmap, gfx_display_columns * 2);		
			}
			walk_display += skipInc;
			
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}

  ta->finish_frame();

  ta->begin_render();
  {
  	vertex_ot_t	v;
  	poly_hdr_t	hdr;
    float fade=1.0f;
    float ratio_h=(float)bitmap->height/((480.0f*(float)gfx_height)/480.0f);
    float ratio_w=(float)bitmap->width/((640.0f*(float)gfx_width)/640.0f);

  	ta->poly_hdr_txr(&hdr, TA_OPAQUE, TA_RGB565, gfx_width, gfx_height, mame_txr, /*TA_NO_FILTER*/TA_BILINEAR_FILTER);
  	ta->commit_poly_hdr(&hdr);

  	/* - -
  	   + - */
  	v.flags = TA_VERTEX_NORMAL;
  	v.x = 0.0f;
  	v.y = 480.0f;
  	v.z = 1.0f;
  	v.u = 0.0f; v.v = ratio_h; 
  	v.a = 1.0f; v.r = v.g = v.b = fade;
  	v.oa = v.or = v.og = v.ob = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* + -
  	   - - */
  	v.y = 0.0f;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* - -
  	   - + */
  	v.x = 640.0f;
  	v.y = 480.0f;
  	v.u = ratio_w; v.v = ratio_h; 
  	ta->commit_vertex(&v, sizeof(v));

  	/* - +
  	   - - */
  	v.flags = TA_VERTEX_EOL;
  	v.y = 0.0f;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));
  }
  ta->commit_eol();
  {
    poly_hdr_t	hdr;
    ta->poly_hdr_col(&hdr, TA_TRANSLUCENT);
    ta->commit_poly_hdr(&hdr);
  }
  ta->commit_eol();
}

void update_screen_dirty_ta_640x480_16_scaled(struct osd_bitmap *bitmap)
{
	register unsigned short * walk_display;
	register unsigned short * walk_bitmap;
	register int i;
	register int j;
	int startRow;
	int endRow;
	int startColumn;
	int endColumn;
	int displayInc;
	int skipInc;

	current_display = ta->txr_map(mame_txr);

	startRow = skiplines;
	endRow = skiplines + gfx_display_lines;
	startColumn = skipcolumns;
	endColumn = skipcolumns + gfx_display_columns;
	
	walk_display = current_display; // + (gfx_xoffset/4) + (gfx_width * (gfx_yoffset / 2));
	displayInc = gfx_width - gfx_display_columns;
	
	walk_bitmap = bitmap->line[startRow] + skipcolumns;
	skipInc = gfx_width;
	
	if (current_pens_lookup)
	{
		for (i = startRow;i < endRow;i++)
		{
			if (ISDIRTY(i))
			{
				for (j = startColumn; j < endColumn; j++)
				{
					*walk_display++ = current_pens_lookup[*walk_bitmap++];
				}
				
				walk_display += displayInc;
			}
			else
			{
				walk_display += skipInc;
			}
			
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}
	else
	{
		for (i = startRow;i < endRow;i++)
		{
			if (ISDIRTY(i))
			{
				mymemcpy(walk_display, walk_bitmap, gfx_display_columns * 2);		
			}
			walk_display += skipInc;
			
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}

  ta->finish_frame();

  ta->begin_render();
  {
  	vertex_ot_t	v;
  	poly_hdr_t	hdr;
    float fade=1.0f;
    float ratio_h;
    float ratio_w;
    float adjust_x;
    float adjust_y;

	if (bitmap->height > bitmap->width)
	{
		ratio_h=(float)bitmap->height/(float)gfx_height;
		ratio_w=ratio_h;//(float)gfx_width/(float)640.0f;
		adjust_x = (gfx_width - bitmap->width)/4;
		adjust_y = 0.0f;
	}
	else
	{
		ratio_w=(float)bitmap->width/(float)gfx_width;
		ratio_h=ratio_w;//(float)gfx_height/(float)480.0f;
		adjust_x = 0.0f;
		adjust_y = (gfx_height - bitmap->height)/4;
	}

  	ta->poly_hdr_txr(&hdr, TA_OPAQUE, TA_RGB565, gfx_width, gfx_height, mame_txr, /*TA_NO_FILTER*/TA_BILINEAR_FILTER);
  	ta->commit_poly_hdr(&hdr);

  	/* - -
  	   + - */
  	v.flags = TA_VERTEX_NORMAL;
  	v.x = 0.0f + adjust_x;
  	v.y = 480.0f + adjust_y;
  	v.z = 1.0f;
  	v.u = 0.0f; v.v = ratio_h; 
  	v.a = 1.0f; v.r = v.g = v.b = fade;
  	v.oa = v.or = v.og = v.ob = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* + -
  	   - - */
  	v.y = 0.0f + adjust_y;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* - -
  	   - + */
  	v.x = 640.0f + adjust_x;
  	v.y = 480.0f + adjust_y;
  	v.u = ratio_w; v.v = ratio_h; 
  	ta->commit_vertex(&v, sizeof(v));

  	/* - +
  	   - - */
  	v.flags = TA_VERTEX_EOL;
  	v.y = 0.0f + adjust_y;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));
  }
  ta->commit_eol();
  {
    poly_hdr_t	hdr;
    ta->poly_hdr_col(&hdr, TA_TRANSLUCENT);
    ta->commit_poly_hdr(&hdr);
  }
  ta->commit_eol();
}

void update_screen_nodirty_8(struct osd_bitmap *bitmap)
{
	register unsigned short * walk_display;
	register unsigned char * walk_bitmap;
	register int i;
	register int j;
	int startRow;
	int endRow;
	int startColumn;
	int endColumn;
	int displayInc;

	current_display = video->vram_s;

	startRow = skiplines;
	endRow = skiplines + gfx_display_lines;
	startColumn = skipcolumns;
	endColumn = skipcolumns + gfx_display_columns;
	
	walk_display = current_display + gfx_xoffset + (gfx_width * gfx_yoffset);
	displayInc = gfx_width - gfx_display_columns;
	
	walk_bitmap = bitmap->line[startRow] + skipcolumns;
	
	if (current_pens_lookup)
	{
		for (i = startRow;i < endRow;i++)
		{
			for (j = startColumn; j < endColumn; j++)
			{
				*walk_display++ = current_pens_lookup[*walk_bitmap++];
			}
			
			walk_display += displayInc;
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}
}

void update_screen_dirty_8(struct osd_bitmap *bitmap)
{
	register unsigned short * walk_display;
	register unsigned char * walk_bitmap;
	register int i;
	register int j;
	int startRow;
	int endRow;
	int startColumn;
	int endColumn;
	int displayInc;
	int skipInc;

	current_display = video->vram_s;

	startRow = skiplines;
	endRow = skiplines + gfx_display_lines;
	startColumn = skipcolumns;
	endColumn = skipcolumns + gfx_display_columns;
	
	walk_display = current_display + gfx_xoffset + (gfx_width * gfx_yoffset);
	displayInc = gfx_width - gfx_display_columns;
	skipInc = gfx_width;
	
	walk_bitmap = bitmap->line[startRow] + skipcolumns;
	
	if (current_pens_lookup)
	{
		for (i = startRow;i < endRow;i++)
		{
			if (ISDIRTY(i))
			{
				for (j = startColumn; j < endColumn; j++)
				{
					*walk_display++ = current_pens_lookup[*walk_bitmap++];
				}
				
				walk_display += displayInc;
			}
			else
			{
				walk_display += skipInc;
			}
			
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}
}

void update_screen_nodirty_ta_320x240_8(struct osd_bitmap *bitmap)
{
	register unsigned short * walk_display;
	register unsigned char * walk_bitmap;
	register int i;
	register int j;
	int startRow;
	int endRow;
	int startColumn;
	int endColumn;
	int displayInc;

	current_display = ta->txr_map(mame_txr);

	startRow = skiplines;
	endRow = skiplines + gfx_display_lines;
	startColumn = skipcolumns;
	endColumn = skipcolumns + gfx_display_columns;
	
	walk_display = current_display; // + (gfx_xoffset/4) + (gfx_width * (gfx_yoffset / 2));
	displayInc = gfx_width - gfx_display_columns;
	
	walk_bitmap = bitmap->line[startRow] + skipcolumns;
	
	if (current_pens_lookup)
	{
		for (i = startRow;i < endRow;i++)
		{
			for (j = startColumn; j < endColumn; j++)
			{
				*walk_display++ = current_pens_lookup[*walk_bitmap++];
			}
			
			walk_display += displayInc;
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}

  ta->finish_frame();

  ta->begin_render();
  {
  	vertex_ot_t	v;
  	poly_hdr_t	hdr;
    float fade=1.0f;
    float ratio_h=(float)bitmap->height/((240.0f*(float)gfx_height)/240.0f);
    float ratio_w=(float)bitmap->width/((320.0f*(float)gfx_width)/320.0f);

  	ta->poly_hdr_txr(&hdr, TA_OPAQUE, TA_RGB565, gfx_width, gfx_height, mame_txr, /*TA_NO_FILTER*/TA_BILINEAR_FILTER);
  	ta->commit_poly_hdr(&hdr);

  	/* - -
  	   + - */
  	v.flags = TA_VERTEX_NORMAL;
  	v.x = 0.0f;
  	v.y = 240.0f;
  	v.z = 1.0f;
  	v.u = 0.0f; v.v = ratio_h; 
  	v.a = 1.0f; v.r = v.g = v.b = fade;
  	v.oa = v.or = v.og = v.ob = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* + -
  	   - - */
  	v.y = 0.0f;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* - -
  	   - + */
  	v.x = 320.0f;
  	v.y = 240.0f;
  	v.u = ratio_w; v.v = ratio_h; 
  	ta->commit_vertex(&v, sizeof(v));

  	/* - +
  	   - - */
  	v.flags = TA_VERTEX_EOL;
  	v.y = 0.0f;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));
  }
  ta->commit_eol();
  {
    poly_hdr_t	hdr;
    ta->poly_hdr_col(&hdr, TA_TRANSLUCENT);
    ta->commit_poly_hdr(&hdr);
  }
  ta->commit_eol();
}

void update_screen_nodirty_ta_640x480_8(struct osd_bitmap *bitmap)
{
	register unsigned short * walk_display;
	register unsigned char * walk_bitmap;
	register int i;
	register int j;
	int startRow;
	int endRow;
	int startColumn;
	int endColumn;
	int displayInc;

	current_display = ta->txr_map(mame_txr);

	startRow = skiplines;
	endRow = skiplines + gfx_display_lines;
	startColumn = skipcolumns;
	endColumn = skipcolumns + gfx_display_columns;
	
	walk_display = current_display; // + (gfx_xoffset/4) + (gfx_width * (gfx_yoffset / 2));
	displayInc = gfx_width - gfx_display_columns;
	
	walk_bitmap = bitmap->line[startRow] + skipcolumns;
	
	if (current_pens_lookup)
	{
		for (i = startRow;i < endRow;i++)
		{
			for (j = startColumn; j < endColumn; j++)
			{
				*walk_display++ = current_pens_lookup[*walk_bitmap++];
			}
			
			walk_display += displayInc;
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}

  ta->finish_frame();

  ta->begin_render();
  {
  	vertex_ot_t	v;
  	poly_hdr_t	hdr;
    float fade=1.0f;
    float ratio_h=(float)bitmap->height/((480.0f*(float)gfx_height)/480.0f);
    float ratio_w=(float)bitmap->width/((640.0f*(float)gfx_width)/640.0f);

  	ta->poly_hdr_txr(&hdr, TA_OPAQUE, TA_RGB565, gfx_width, gfx_height, mame_txr, /*TA_NO_FILTER*/TA_BILINEAR_FILTER);
  	ta->commit_poly_hdr(&hdr);

  	/* - -
  	   + - */
  	v.flags = TA_VERTEX_NORMAL;
  	v.x = 0.0f;
  	v.y = 480.0f;
  	v.z = 1.0f;
  	v.u = 0.0f; v.v = ratio_h; 
  	v.a = 1.0f; v.r = v.g = v.b = fade;
  	v.oa = v.or = v.og = v.ob = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* + -
  	   - - */
  	v.y = 0.0f;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* - -
  	   - + */
  	v.x = 640.0f;
  	v.y = 480.0f;
  	v.u = ratio_w; v.v = ratio_h; 
  	ta->commit_vertex(&v, sizeof(v));

  	/* - +
  	   - - */
  	v.flags = TA_VERTEX_EOL;
  	v.y = 0.0f;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));
  }
  ta->commit_eol();
  {
    poly_hdr_t	hdr;
    ta->poly_hdr_col(&hdr, TA_TRANSLUCENT);
    ta->commit_poly_hdr(&hdr);
  }
  ta->commit_eol();
}

void update_screen_nodirty_ta_640x480_8_scaled(struct osd_bitmap *bitmap)
{
	register unsigned short * walk_display;
	register unsigned char * walk_bitmap;
	register int i;
	register int j;
	int startRow;
	int endRow;
	int startColumn;
	int endColumn;
	int displayInc;

	current_display = ta->txr_map(mame_txr);

	startRow = skiplines;
	endRow = skiplines + gfx_display_lines;
	startColumn = skipcolumns;
	endColumn = skipcolumns + gfx_display_columns;
	
	walk_display = current_display; // + (gfx_xoffset/4) + (gfx_width * (gfx_yoffset / 2));
	displayInc = gfx_width - gfx_display_columns;
	
	walk_bitmap = bitmap->line[startRow] + skipcolumns;
	
	if (current_pens_lookup)
	{
		for (i = startRow;i < endRow;i++)
		{
			for (j = startColumn; j < endColumn; j++)
			{
				*walk_display++ = current_pens_lookup[*walk_bitmap++];
			}
			
			walk_display += displayInc;
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}

  ta->finish_frame();

  ta->begin_render();
  {
  	vertex_ot_t	v;
  	poly_hdr_t	hdr;
    float fade=1.0f;
    float ratio_h;
    float ratio_w;
    float adjust_x;
    float adjust_y;

	if (bitmap->height > bitmap->width)
	{
		ratio_h=(float)bitmap->height/(float)gfx_height;
		ratio_w=ratio_h;//(float)gfx_width/(float)640.0f;
		adjust_x = (gfx_width - bitmap->width)/4;
		adjust_y = 0.0f;
	}
	else
	{
		ratio_w=(float)bitmap->width/(float)gfx_width;
		ratio_h=ratio_w;//(float)gfx_height/(float)480.0f;
		adjust_x = 0.0f;
		adjust_y = (gfx_height - bitmap->height)/4;
	}
	
  	ta->poly_hdr_txr(&hdr, TA_OPAQUE, TA_RGB565, gfx_width, gfx_height, mame_txr, /*TA_NO_FILTER*/TA_BILINEAR_FILTER);
  	ta->commit_poly_hdr(&hdr);

  	/* - -
  	   + - */
  	v.flags = TA_VERTEX_NORMAL;
  	v.x = 0.0f + adjust_x;
  	v.y = 480.0f + adjust_y;
  	v.z = 1.0f;
  	v.u = 0.0f; v.v = ratio_h; 
  	v.a = 1.0f; v.r = v.g = v.b = fade;
  	v.oa = v.or = v.og = v.ob = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* + -
  	   - - */
  	v.y = 0.0f + adjust_y;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* - -
  	   - + */
  	v.x = 640.0f + adjust_x;
  	v.y = 480.0f + adjust_y;
  	v.u = ratio_w; v.v = ratio_h; 
  	ta->commit_vertex(&v, sizeof(v));

  	/* - +
  	   - - */
  	v.flags = TA_VERTEX_EOL;
  	v.y = 0.0f + adjust_y;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));
  }
  ta->commit_eol();
  {
    poly_hdr_t	hdr;
    ta->poly_hdr_col(&hdr, TA_TRANSLUCENT);
    ta->commit_poly_hdr(&hdr);
  }
  ta->commit_eol();
}

void update_screen_dirty_ta_320x240_8(struct osd_bitmap *bitmap)
{
	register unsigned short * walk_display;
	register unsigned char * walk_bitmap;
	register int i;
	register int j;
	int startRow;
	int endRow;
	int startColumn;
	int endColumn;
	int displayInc;
	int skipInc;

	current_display = ta->txr_map(mame_txr);

	startRow = skiplines;
	endRow = skiplines + gfx_display_lines;
	startColumn = skipcolumns;
	endColumn = skipcolumns + gfx_display_columns;
	
	walk_display = current_display; // + (gfx_xoffset/4) + (gfx_width * (gfx_yoffset / 2));
	displayInc = gfx_width - gfx_display_columns;
	
	walk_bitmap = bitmap->line[startRow] + skipcolumns;
	skipInc = gfx_width;
	
	if (current_pens_lookup)
	{
		for (i = startRow;i < endRow;i++)
		{
			if (ISDIRTY(i))
			{
				for (j = startColumn; j < endColumn; j++)
				{
					*walk_display++ = current_pens_lookup[*walk_bitmap++];
				}
				
				walk_display += displayInc;
			}
			else
			{
				walk_display += skipInc;
			}
			
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}

  ta->finish_frame();

  ta->begin_render();
  {
  	vertex_ot_t	v;
  	poly_hdr_t	hdr;
    float fade=1.0f;
    float ratio_h=(float)bitmap->height/((240.0f*(float)gfx_height)/240.0f);
    float ratio_w=(float)bitmap->width/((320.0f*(float)gfx_width)/320.0f);

  	ta->poly_hdr_txr(&hdr, TA_OPAQUE, TA_RGB565, gfx_width, gfx_height, mame_txr, /*TA_NO_FILTER*/TA_BILINEAR_FILTER);
  	ta->commit_poly_hdr(&hdr);

  	/* - -
  	   + - */
  	v.flags = TA_VERTEX_NORMAL;
  	v.x = 0.0f;
  	v.y = 240.0f;
  	v.z = 1.0f;
  	v.u = 0.0f; v.v = ratio_h; 
  	v.a = 1.0f; v.r = v.g = v.b = fade;
  	v.oa = v.or = v.og = v.ob = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* + -
  	   - - */
  	v.y = 0.0f;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* - -
  	   - + */
  	v.x = 320.0f;
  	v.y = 240.0f;
  	v.u = ratio_w; v.v = ratio_h; 
  	ta->commit_vertex(&v, sizeof(v));

  	/* - +
  	   - - */
  	v.flags = TA_VERTEX_EOL;
  	v.y = 0.0f;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));
  }
  ta->commit_eol();
  {
    poly_hdr_t	hdr;
    ta->poly_hdr_col(&hdr, TA_TRANSLUCENT);
    ta->commit_poly_hdr(&hdr);
  }
  ta->commit_eol();
}

void update_screen_dirty_ta_640x480_8(struct osd_bitmap *bitmap)
{
	register unsigned short * walk_display;
	register unsigned char * walk_bitmap;
	register int i;
	register int j;
	int startRow;
	int endRow;
	int startColumn;
	int endColumn;
	int displayInc;
	int skipInc;

	current_display = ta->txr_map(mame_txr);

	startRow = skiplines;
	endRow = skiplines + gfx_display_lines;
	startColumn = skipcolumns;
	endColumn = skipcolumns + gfx_display_columns;
	
	walk_display = current_display; // + (gfx_xoffset/4) + (gfx_width * (gfx_yoffset / 2));
	displayInc = gfx_width - gfx_display_columns;
	
	walk_bitmap = bitmap->line[startRow] + skipcolumns;
	skipInc = gfx_width;
	
	if (current_pens_lookup)
	{
		for (i = startRow;i < endRow;i++)
		{
			if (ISDIRTY(i))
			{
				for (j = startColumn; j < endColumn; j++)
				{
					*walk_display++ = current_pens_lookup[*walk_bitmap++];
				}
				
				walk_display += displayInc;
			}
			else
			{
				walk_display += skipInc;
			}
			
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}

  ta->finish_frame();

  ta->begin_render();
  {
  	vertex_ot_t	v;
  	poly_hdr_t	hdr;
    float fade=1.0f;
    float ratio_h=(float)bitmap->height/((480.0f*(float)gfx_height)/480.0f);
    float ratio_w=(float)bitmap->width/((640.0f*(float)gfx_width)/640.0f);

  	ta->poly_hdr_txr(&hdr, TA_OPAQUE, TA_RGB565, gfx_width, gfx_height, mame_txr, /*TA_NO_FILTER*/TA_BILINEAR_FILTER);
  	ta->commit_poly_hdr(&hdr);

  	/* - -
  	   + - */
  	v.flags = TA_VERTEX_NORMAL;
  	v.x = 0.0f;
  	v.y = 480.0f;
  	v.z = 1.0f;
  	v.u = 0.0f; v.v = ratio_h; 
  	v.a = 1.0f; v.r = v.g = v.b = fade;
  	v.oa = v.or = v.og = v.ob = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* + -
  	   - - */
  	v.y = 0.0f;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* - -
  	   - + */
  	v.x = 640.0f;
  	v.y = 480.0f;
  	v.u = ratio_w; v.v = ratio_h; 
  	ta->commit_vertex(&v, sizeof(v));

  	/* - +
  	   - - */
  	v.flags = TA_VERTEX_EOL;
  	v.y = 0.0f;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));
  }
  ta->commit_eol();
  {
    poly_hdr_t	hdr;
    ta->poly_hdr_col(&hdr, TA_TRANSLUCENT);
    ta->commit_poly_hdr(&hdr);
  }
  ta->commit_eol();
}

void update_screen_dirty_ta_640x480_8_scaled(struct osd_bitmap *bitmap)
{
	register unsigned short * walk_display;
	register unsigned char * walk_bitmap;
	register int i;
	register int j;
	int startRow;
	int endRow;
	int startColumn;
	int endColumn;
	int displayInc;
	int skipInc;

	current_display = ta->txr_map(mame_txr);

	startRow = skiplines;
	endRow = skiplines + gfx_display_lines;
	startColumn = skipcolumns;
	endColumn = skipcolumns + gfx_display_columns;
	
	walk_display = current_display; // + (gfx_xoffset/4) + (gfx_width * (gfx_yoffset / 2));
	displayInc = gfx_width - gfx_display_columns;
	
	walk_bitmap = bitmap->line[startRow] + skipcolumns;
	skipInc = gfx_width;
	
	if (current_pens_lookup)
	{
		for (i = startRow;i < endRow;i++)
		{
			if (ISDIRTY(i))
			{
				for (j = startColumn; j < endColumn; j++)
				{
					*walk_display++ = current_pens_lookup[*walk_bitmap++];
				}
				
				walk_display += displayInc;
			}
			else
			{
				walk_display += skipInc;
			}
			
			walk_bitmap = bitmap->line[i] + skipcolumns;
		}
	}

  ta->finish_frame();

  ta->begin_render();
  {
  	vertex_ot_t	v;
  	poly_hdr_t	hdr;
    float fade=1.0f;
    float ratio_h;
    float ratio_w;
    float adjust_x;
    float adjust_y;

	if (bitmap->height > bitmap->width)
	{
		ratio_h=(float)bitmap->height/(float)gfx_height;
		ratio_w=ratio_h;//(float)gfx_width/(float)640.0f;
		adjust_x = (gfx_width - bitmap->width)/4;
		adjust_y = 0.0f;
	}
	else
	{
		ratio_w=(float)bitmap->width/(float)gfx_width;
		ratio_h=ratio_w;//(float)gfx_height/(float)480.0f;
		adjust_x = 0.0f;
		adjust_y = (gfx_height - bitmap->height)/4;
	}

  	ta->poly_hdr_txr(&hdr, TA_OPAQUE, TA_RGB565, gfx_width, gfx_height, mame_txr, /*TA_NO_FILTER*/TA_BILINEAR_FILTER);
  	ta->commit_poly_hdr(&hdr);

  	/* - -
  	   + - */
  	v.flags = TA_VERTEX_NORMAL;
  	v.x = 0.0f + adjust_x;
  	v.y = 480.0f + adjust_y;
  	v.z = 1.0f;
  	v.u = 0.0f; v.v = ratio_h; 
  	v.a = 1.0f; v.r = v.g = v.b = fade;
  	v.oa = v.or = v.og = v.ob = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* + -
  	   - - */
  	v.y = 0.0f + adjust_y;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));

  	/* - -
  	   - + */
  	v.x = 640.0f + adjust_x;
  	v.y = 480.0f + adjust_y;
  	v.u = ratio_w; v.v = ratio_h; 
  	ta->commit_vertex(&v, sizeof(v));

  	/* - +
  	   - - */
  	v.flags = TA_VERTEX_EOL;
  	v.y = 0.0f + adjust_y;
  	v.v = 0.0f;
  	ta->commit_vertex(&v, sizeof(v));
  }
  ta->commit_eol();
  {
    poly_hdr_t	hdr;
    ta->poly_hdr_col(&hdr, TA_TRANSLUCENT);
    ta->commit_poly_hdr(&hdr);
  }
  ta->commit_eol();
}

struct osd_bitmap *osd_alloc_bitmap(int width,int height,int depth)
{
	struct osd_bitmap *bitmap;

//	printf("osd_alloc_bitmap(%d, %d, %d)\n", width, height, depth);
	
	if (depth != 8 && depth != 15 && depth != 16 && depth != 32)
	{
		logerror("osd_alloc_bitmap() unknown depth %d\n",depth);
		return NULL;
	}

	// debug requires at least this
	if (options.mame_debug)
	{
		width = 640;
		height = 480;
	}
	
	if ((bitmap = malloc(sizeof(struct osd_bitmap))) != 0)
	{
		int i,rowlen,rdwidth;
		unsigned char *bm;


		bitmap->depth = depth;
		bitmap->width = width;
		bitmap->height = height;
		
		rdwidth = (width + 7) & ~7;     /* round width to a quadword */
		rowlen = (rdwidth + 2 * safety) * sizeof(unsigned char);
		if (depth == 32)
			rowlen *= 4;
		else if (depth == 15 || depth == 16)
			rowlen *= 2;

		if ((bm = malloc((height + 2 * safety) * rowlen)) == 0)
		{
			free(bitmap);
			return 0;
		}

		/* clear ALL bitmap, including safety area, to avoid garbage on right */
		/* side of screen if width is not a multiple of 4 */
		memset(bm,0,(height + 2 * safety) * rowlen);

		if ((bitmap->line = malloc((height + 2 * safety) * sizeof(unsigned char *))) == 0)
		{
			free(bm);
			free(bitmap);
			return 0;
		}

		for (i = 0;i < height + 2 * safety;i++)
		{
			if (depth == 32)
				bitmap->line[i] = &bm[i * rowlen + 4*safety];
			else if (depth == 15 || depth == 16)
				bitmap->line[i] = &bm[i * rowlen + 2*safety];
			else
				bitmap->line[i] = &bm[i * rowlen + safety];
		}
		bitmap->line += safety;
		
		bitmap->_private = bm;
	}
	
	return bitmap;
}


#if 1
/* set the bitmap to black */
void osd_clearbitmap(struct osd_bitmap *bitmap)
{
	int i;


//	printf("osd_clearbitmap(%d)\n", bitmap);

	for (i = 0;i < bitmap->height;i++)
	{
		if (bitmap->depth == 16)
			memset(bitmap->line[i],0,2*bitmap->width);
		else
			memset(bitmap->line[i],BACKGROUND,bitmap->width);
	}


	if (bitmap == Machine->scrbitmap || bitmap == artwork_real_scrbitmap)
	{
		extern int bitmap_dirty;        /* in mame.c */

		osd_mark_dirty (0,0,bitmap->width-1,bitmap->height-1);
		bitmap_dirty = 1;
	}
}
#endif


void osd_free_bitmap(struct osd_bitmap *bitmap)
{
//	printf("osd_free_bitmap(%d)\n", bitmap);

	if (bitmap)
	{
		bitmap->line -= safety;
		free(bitmap->line);
		free(bitmap->_private);
		free(bitmap);
	}
}


void osd_mark_dirty(int _x1, int _y1, int _x2, int _y2)
{
//	printf("osd_mark_dirty(%d,%d,%d,%d)\n", _x1, _y1, _x2, _y2);

	if (use_dirty)
	{
		int y;
		
//        logerror("mark_dirty %3d,%3d - %3d,%3d\n", _x1,_y1, _x2,_y2);

//		_y1 -= skiplines;
//		_y2 -= skiplines;

//	if (_y1 >= gfx_display_lines || _y2 < 0 || _x1 > gfx_display_columns || _x2 < 0) return;
		if (_y1 < skiplines) _y1 = 0;
		if (_y2 >= (gfx_display_lines + skiplines)) _y2 = gfx_display_lines + skiplines - 1;
//		if (_x1 < 0) _x1 = 0;
//		if (_x2 >= gfx_width) _x2 = gfx_width - 1;

/* Digita, vector games actually need one extra pixel dirty */
		for (y = _y1; y <= _y2 + 2; y++)
//			for (x = _x1; x <= _x2 + 15; x += 16)
				MARKDIRTY(y);
	}
}

static void init_dirty(char dirty)
{
	memset(dirty_new, dirty, MAX_GFX_HEIGHT);
}

INLINE void swap_dirty(void)
{
    char *tmp;

	tmp = dirty_old;
	dirty_old = dirty_new;
	dirty_new = tmp;
}



/*
 * This function tries to find the best display mode.
 */
static void select_display_mode(int width,int height,int depth,int attributes,int orientation)
{
	switch (gResolution)
	{
		case 0:
		{
			gfx_width = 320;
			gfx_height = 240;

			video->set_mode(video->check_cable(), DM_320x240, PM_RGB565);
			video->clear(0,0,0);
			
			if (options.color_depth == 8)
			{
				if (use_dirty)
				{
					update_screen = update_screen_dirty_8;
				}
				else
				{
					update_screen = update_screen_nodirty_8;
				}
			}
			else
			{
				if (use_dirty)
				{
					update_screen = update_screen_dirty_16;
				}
				else
				{
					update_screen = update_screen_nodirty_16;
				}
			}
			// set this to our size so we don't try and go over
			options.vector_width = gfx_width;
			options.vector_height = gfx_height;
			break;
		}
		case 3:
		{
			gfx_width = 640;
			gfx_height = 480;

			video->set_mode(video->check_cable(), DM_640x480, PM_RGB565);
			video->clear(0,0,0);
			
			if (options.color_depth == 8)
			{
				if (use_dirty)
				{
					update_screen = update_screen_dirty_8;
				}
				else
				{
					update_screen = update_screen_nodirty_8;
				}
			}
			else
			{
				if (use_dirty)
				{
					update_screen = update_screen_dirty_16;
				}
				else
				{
					update_screen = update_screen_nodirty_16;
				}
			}
			
			// set this to our size so we don't try and go over
			options.vector_width = gfx_width;
			options.vector_height = gfx_height;
			break;
		}
		case 1:
		{
			video->set_mode(video->check_cable(), DM_640x480, PM_RGB565);
			video->clear(0,0,0);

			ta->hw_init();
			
#define MIN_TA_SIZE 256

			if (options.color_depth == 8)
			{
				if (use_dirty)
				{
					update_screen = update_screen_dirty_ta_640x480_8;
				}
				else
				{
					update_screen = update_screen_nodirty_ta_640x480_8;
				}
			}
			else
			{
				if (use_dirty)
				{
					update_screen = update_screen_dirty_ta_640x480_16;
				}
				else
				{
					update_screen = update_screen_nodirty_ta_640x480_16;
				}
			}
			
			gfx_width = (width / MIN_TA_SIZE) * MIN_TA_SIZE;
			gfx_height = (height / MIN_TA_SIZE) * MIN_TA_SIZE;

			if (width > gfx_width) 
				gfx_width += MIN_TA_SIZE;
			if (height > gfx_height) 
				gfx_height+=MIN_TA_SIZE;

			if (gfx_width != gfx_height)
			{
				if (gfx_width > gfx_height)
					gfx_height = gfx_width;
				else
					gfx_width = gfx_height;
			}
			mame_txr = ta->txr_allocate(gfx_width * gfx_height * 2);
			break;
		}
		case -1:
		{
			video->set_mode(video->check_cable(), DM_640x480, PM_RGB565);
			video->clear(0,0,0);

			ta->hw_init();
			
#define MIN_TA_SIZE 256

			if (options.color_depth == 8)
			{
				if (use_dirty)
				{
					update_screen = update_screen_dirty_ta_640x480_8_scaled;
				}
				else
				{
					update_screen = update_screen_nodirty_ta_640x480_8_scaled;
				}
			}
			else
			{
				if (use_dirty)
				{
					update_screen = update_screen_dirty_ta_640x480_16_scaled;
				}
				else
				{
					update_screen = update_screen_nodirty_ta_640x480_16_scaled;
				}
			}
			
			gfx_width = (width / MIN_TA_SIZE) * MIN_TA_SIZE;
			gfx_height = (height / MIN_TA_SIZE) * MIN_TA_SIZE;

			if (width > gfx_width) 
				gfx_width += MIN_TA_SIZE;
			if (height > gfx_height) 
				gfx_height+=MIN_TA_SIZE;

			if (gfx_width != gfx_height)
			{
				if (gfx_width > gfx_height)
					gfx_height = gfx_width;
				else
					gfx_width = gfx_height;
			}
			mame_txr = ta->txr_allocate(gfx_width * gfx_height * 2);
			break;
		}
		case 2:
		{
			video->set_mode(video->check_cable(), DM_320x240, PM_RGB565);
			video->clear(0,0,0);

			ta->hw_init();
			
#define MIN_TA_SIZE 256

			if (options.color_depth == 8)
			{
				if (use_dirty)
				{
					update_screen = update_screen_dirty_ta_320x240_8;
				}
				else
				{
					update_screen = update_screen_nodirty_ta_320x240_8;
				}
			}
			else
			{
				if (use_dirty)
				{
					update_screen = update_screen_dirty_ta_320x240_16;
				}
				else
				{
					update_screen = update_screen_nodirty_ta_320x240_16;
				}
			}
			
			gfx_width = (width / MIN_TA_SIZE) * MIN_TA_SIZE;
			gfx_height = (height / MIN_TA_SIZE) * MIN_TA_SIZE;

			if (width > gfx_width) 
				gfx_width += MIN_TA_SIZE;
			if (height > gfx_height) 
				gfx_height+=MIN_TA_SIZE;

			if (gfx_width != gfx_height)
			{
				if (gfx_width > gfx_height)
					gfx_height = gfx_width;
				else
					gfx_width = gfx_height;
			}
			mame_txr = ta->txr_allocate(gfx_width * gfx_height * 2);
			break;
		}
		default:
		{
			printf("Unknown display resolution\n");
			break;
		}
	}
}



/* center image inside the display based on the visual area */
static void internal_set_visible_area(int min_x,int max_x,int min_y,int max_y,int debugger)
{
	int act_width;

logerror("set visible area %d-%d %d-%d\n",min_x,max_x,min_y,max_y);

		act_width = gfx_width;


	viswidth  = max_x - min_x + 1;
	visheight = max_y - min_y + 1;

		xmultiply = ymultiply = 1;

	gfx_display_lines = visheight;
	gfx_display_columns = viswidth;

	gfx_xoffset = (act_width - viswidth * xmultiply) / 2;
	if (gfx_display_columns > act_width / xmultiply)
		gfx_display_columns = act_width / xmultiply;

	gfx_yoffset = (gfx_height - visheight * ymultiply) / 2;
		if (gfx_display_lines > gfx_height / ymultiply)
			gfx_display_lines = gfx_height / ymultiply;


	skiplinesmin = min_y;
	skiplinesmax = visheight - gfx_display_lines + min_y;
	skipcolumnsmin = min_x;
	skipcolumnsmax = viswidth - gfx_display_columns + min_x;

	/* Align on a quadword !*/
	gfx_xoffset &= ~7;

	/* the skipcolumns from mame.cfg/cmdline is relative to the visible area */
	skipcolumns = min_x + skipcolumns;
	skiplines   = min_y + skiplines;

	/* Just in case the visual area doesn't fit */
	if (gfx_xoffset < 0)
	{
		skipcolumns -= gfx_xoffset;
		gfx_xoffset = 0;
	}
	if (gfx_yoffset < 0)
	{
		skiplines   -= gfx_yoffset;
		gfx_yoffset = 0;
	}

	/* Failsafe against silly parameters */
	if (skiplines < skiplinesmin)
		skiplines = skiplinesmin;
	if (skipcolumns < skipcolumnsmin)
		skipcolumns = skipcolumnsmin;
	if (skiplines > skiplinesmax)
		skiplines = skiplinesmax;
	if (skipcolumns > skipcolumnsmax)
		skipcolumns = skipcolumnsmax;

	logerror("gfx_width = %d gfx_height = %d\n"
				"gfx_xoffset = %d gfx_yoffset = %d\n"
				"xmin %d ymin %d xmax %d ymax %d\n"
				"skiplines %d skipcolumns %d\n"
				"gfx_display_lines %d gfx_display_columns %d\n"
				"xmultiply %d ymultiply %d\n",
				gfx_width,gfx_height,
				gfx_xoffset,gfx_yoffset,
				min_x, min_y, max_x, max_y, skiplines, skipcolumns,gfx_display_lines,gfx_display_columns,xmultiply,ymultiply);

	set_ui_visarea(skipcolumns, skiplines, skipcolumns+gfx_display_columns-1, skiplines+gfx_display_lines-1);

	/* round to a multiple of 4 to avoid missing pixels on the right side */
	gfx_display_columns  = (gfx_display_columns + 3) & ~3;
}


void osd_set_visible_area(int min_x,int max_x,int min_y,int max_y)
{
//	printf("osd_set_visible_area(%d,%d,%d,%d)\n", min_x, max_x, min_y, max_y);

	vis_min_x = min_x;
	vis_max_x = max_x;
	vis_min_y = min_y;
	vis_max_y = max_y;
	internal_set_visible_area(min_x,max_x,min_y,max_y,0);
}




/* set the actual display screen but don't allocate the screen bitmap */
static int osd_set_display(int width,int height,int depth,int attributes,int orientation)
{
	int     i;
	
//	printf("osd_set_display(%d,%d,%d,%d,%d)\n", width, height, depth, attributes, orientation);

	if (!gfx_height || !gfx_width)
	{
		printf("Please specify height AND width (e.g. -640x480)\n");
		return 0;
	}


	/* Mark the dirty buffers as dirty */

	if (use_dirty)
	{
		if (vector_game)
			/* vector games only use one dirty buffer */
			init_dirty (0);
		else
			init_dirty(1);
		swap_dirty();
		init_dirty(1);
	}
	if (dirtycolor)
	{
		for (i = 0;i < screen_colors;i++)
			dirtycolor[i] = 1;
		dirtypalette = 1;
	}


	gone_to_gfx_mode = 1;

	
	vsync_frame_rate = video_fps;


	warming_up = 1;
	
	return 1;
}


/*
Create a display screen, or window, of the given dimensions (or larger).
Attributes are the ones defined in driver.h.
Returns 0 on success.
*/
int osd_create_display(int width,int height,int depth,int fps,int attributes,int orientation)
{
	logerror("width %d, height %d depth %d\n",width,height,depth);

//	printf("osd_create_display(%d,%d,%d,%d,%d,%d)\n", width, height, depth, fps, attributes, orientation);
	video_depth = depth;
	if (video_depth == 15) video_depth = 16;
	video_fps = fps;
	video_attributes = attributes;
	video_orientation = orientation;

	show_debugger = 0;

	brightness = 100;
	brightness_paused_adjust = 1.0;
	dirty_bright = 1;

	if (frameskip < 0) frameskip = 0;
	if (frameskip >= FRAMESKIP_LEVELS) frameskip = FRAMESKIP_LEVELS-1;
	


	gone_to_gfx_mode = 0;

	/* Look if this is a vector game */
	if (attributes & VIDEO_TYPE_VECTOR)
		vector_game = 1;
	else
		vector_game = 0;


	if (use_dirty == -1)	/* dirty=auto in mame.cfg? */
	{
		/* Is the game using a dirty system? */
		if ((attributes & VIDEO_SUPPORTS_DIRTY) || vector_game)
			use_dirty = 1;
		else
			use_dirty = 0;
	}

	select_display_mode(width,height,video_depth,attributes,orientation);



	if (!osd_set_display(width,height,video_depth,attributes,orientation))
		return 1;

	/* set visible area to nothing just to initialize it - it will be set by the core */
	osd_set_visible_area(0,0,0,0);

    return 0;
}


/* shut up the display */
void osd_close_display(void)
{
//printf("osd_close_display\n");
	if (pens_lookup)
	{
		free(pens_lookup);
		pens_lookup = 0;
	}

	if (debug_pens_lookup)
	{
		free(debug_pens_lookup);
		debug_pens_lookup = 0;
	}

	current_display = 0;

	free(dirtycolor);
	dirtycolor = 0;
	free(current_palette);
	current_palette = 0;
}


void osd_debugger_focus(int debugger_has_focus)
{
//	printf("osd_debugger_focus %d\n", debugger_has_focus);
	
	if (debugger_has_focus)
		current_pens_lookup = debug_pens_lookup;
	else
		current_pens_lookup = pens_lookup;
		
    if (show_debugger != debugger_has_focus)
	{
		int i;
		show_debugger = debugger_has_focus;
		debugger_focus_changed = 1;
		for (i = 0;i < screen_colors;i++)
			dirtycolor[i] = 1;
		dirtypalette = 1;

		if (!show_debugger)
		{
			/* silly way to clear the screen */
			struct osd_bitmap *clrbitmap;
			clrbitmap = osd_alloc_bitmap(gfx_display_columns,gfx_display_lines,video_depth);
			if (clrbitmap)
			{
				fillbitmap(clrbitmap,Machine->uifont->colortable[0],NULL);
				/* three times to handle triple buffering */
				update_screen_debugger(clrbitmap);
				update_screen_debugger(clrbitmap);
				update_screen_debugger(clrbitmap);
				osd_free_bitmap(clrbitmap);
			}
		}
	}
}


int osd_allocate_colors(unsigned int totalcolors,
		const UINT8 *palette,UINT32 *pens,int modifiable,
		const UINT8 *debug_palette,UINT32 *debug_pens)
{
	int i;

	if  (options.color_depth == 8)
	{
		modifiable = 1;
	}
	
	modifiable_palette = modifiable;
	screen_colors = totalcolors;
	
	dirtycolor = malloc(screen_colors * sizeof(int));

	current_palette = malloc(3 * screen_colors * sizeof(unsigned char) + 3);
	if (dirtycolor == 0 || current_palette == 0)
		return 1;

	for (i = 0;i < screen_colors;i++)
		dirtycolor[i] = 1;
	dirtypalette = 1;
	for (i = 0;i < screen_colors;i++)
		current_palette[3*i+0] = current_palette[3*i+1] = current_palette[3*i+2] = 0;

	if (modifiable_palette)
	{
		// add two for B&W at the end
		pens_lookup = malloc((screen_colors + 2) * 2);
		debug_pens_lookup = malloc(DEBUGGER_TOTAL_COLORS * 2);
	}
	
	for (i = 0; i < screen_colors; i++)
	{
		unsigned int r,g,b;
		
		r = palette[3*i+0];
		g = palette[3*i+1];
		b = palette[3*i+2];
		
		current_palette[3*i+0] = r;
		current_palette[3*i+1] = g;
		current_palette[3*i+2] = b;
		
//		r = 255.0 * pow((float)(r) / 255.0, 1.0 / osd_gamma_correction);
//		g = 255.0 * pow((float)(g) / 255.0, 1.0 / osd_gamma_correction);
//		b = 255.0 * pow((float)(b) / 255.0, 1.0 / osd_gamma_correction);

		r = (int)(r * brightness * brightness_paused_adjust / 100.0);
		g = (int)(g * brightness * brightness_paused_adjust / 100.0);
		b = (int)(b * brightness * brightness_paused_adjust / 100.0);
		
		if (pens_lookup)
		{
			pens[i] = i;
			pens_lookup[i] = ((r >> 3) << 11) | ((g >> 2) << 5) | ((b >> 3) << 0);
		}
		else
		{
			pens[i] = ((r >> 3) << 11) | ((g >> 2) << 5) | ((b >> 3) << 0);
		}
	}

	// add B&W at the end but not to the pens[]
	if (pens_lookup)
	{
		pens_lookup[screen_colors] = 0;
		pens_lookup[screen_colors + 1] = 0xFFFF;
	}
	
	if (debug_palette && debug_pens)
	{
		for (i = 0; i < DEBUGGER_TOTAL_COLORS; i++)
		{
			unsigned int r,g,b;
			
			r = debug_palette[3*i+0];
			g = debug_palette[3*i+1];
			b = debug_palette[3*i+2];
	
//		r = 255.0 * pow((float)(r) / 255.0, 1.0 / osd_gamma_correction);
//		g = 255.0 * pow((float)(g) / 255.0, 1.0 / osd_gamma_correction);
//		b = 255.0 * pow((float)(b) / 255.0, 1.0 / osd_gamma_correction);

			r = (int)(r * brightness * brightness_paused_adjust / 100.0);
			g = (int)(g * brightness * brightness_paused_adjust / 100.0);
			b = (int)(b * brightness * brightness_paused_adjust / 100.0);
		
			if (debug_pens_lookup)
			{
				debug_pens[i] = i;
				debug_pens_lookup[i] = ((r >> 3) << 11) | ((g >> 2) << 5) | ((b >> 3) << 0);
			}
			else
			{
				debug_pens[i] = ((r >> 3) << 11) | ((g >> 2) << 5) | ((b >> 3) << 0);
			}
		}
	}
	
	if (pens_lookup)
	{
#if 0
		unsigned int i, total, highest_index = 1, lowest_index = 0, highest = 0, lowest = 65536;
		unsigned int r,g,b;
		
		// quick search to find a close black & white
		for (i = 0; i < screen_colors; i++)
		{
			r = current_palette[3*i+0];
			g = current_palette[3*i+1];
			b = current_palette[3*i+2];
			
			total = r + g + b;
			
			if (highest < total)
			{
				printf("highest %d, rgb %d %d %d\n", i, r,g,b);
				highest_index = i;
				highest = total;
			}
			if (lowest > total)
			{
				printf("lowest %d, rgb %d %d %d\n", i, r,g,b);
				lowest_index = i;
				lowest = total;
			}
		}
		
		Machine->uifont->colortable[0] = lowest_index;
		Machine->uifont->colortable[1] = highest_index;
		Machine->uifont->colortable[2] = highest_index;
		Machine->uifont->colortable[3] = lowest_index;
#endif		
		Machine->uifont->colortable[0] = screen_colors;
		Machine->uifont->colortable[1] = screen_colors + 1;
		Machine->uifont->colortable[2] = screen_colors + 1;
		Machine->uifont->colortable[3] = screen_colors;
		
		
//		current_palette[3*totalcolors+0] = 0;
//		pens_lookup[totalcolors] = 0;
//		pens_lookup[totalcolors + 1] = ((255 >> 3) << 11) | ((255 >> 2) << 5) | ((255 >> 3) << 0);
//		current_palette[3*totalcolors+1] = 0;
//		current_palette[3*totalcolors+2] = 0;
//		current_palette[3*(totalcolors + 1) +0] = 255;
//		current_palette[3*(totalcolors + 1) +1] = 255;
//		current_palette[3*(totalcolors + 1) +2] = 255;
	}
	else
	{
		// we have full control to set our own B&W
		Machine->uifont->colortable[0] = 0;
		Machine->uifont->colortable[1] = ((255 >> 3) << 11) | ((255 >> 2) << 5) | ((255 >> 3) << 0);
		Machine->uifont->colortable[2] = ((255 >> 3) << 11) | ((255 >> 2) << 5) | ((255 >> 3) << 0);
		Machine->uifont->colortable[3] = 0;
	}

	current_pens_lookup = pens_lookup;
	
	return 0;
}


void osd_modify_pen(int pen,unsigned char red, unsigned char green, unsigned char blue)
{
	if (modifiable_palette == 0)
	{
		logerror("error: osd_modify_pen() called with modifiable_palette == 0\n");
		return;
	}


//	printf("osd_modify_pen(%d,%d,%d,%d)\n", pen, red, green, blue);
	if (	current_palette[3*pen+0] != red ||
			current_palette[3*pen+1] != green ||
			current_palette[3*pen+2] != blue)
	{
		current_palette[3*pen+0] = red;
		current_palette[3*pen+1] = green;
		current_palette[3*pen+2] = blue;

		dirtycolor[pen] = 1;
		dirtypalette = 1;
#if 0
		if (current_pens_lookup)
		{
			unsigned int i, total, highest_index = 1, lowest_index = 0, highest = 0, lowest = 65536;
			unsigned int r,g,b;
			
			// quick search to find a close black & white
			for (i = 0; i < screen_colors; i++)
			{
				r = current_palette[3*i+0];
				g = current_palette[3*i+1];
				b = current_palette[3*i+2];
				
				total = r + g + b;
				
				if (highest < total)
				{
					printf("highest %d, rgb %d %d %d\n", i, r,g,b);
					highest_index = i;
					highest = total;
				}
				if (lowest > total)
				{
					printf("lowest %d, rgb %d %d %d\n", i, r,g,b);
					lowest_index = i;
					lowest = total;
				}
			}
			
			Machine->uifont->colortable[0] = lowest_index;
			Machine->uifont->colortable[1] = highest_index;
			Machine->uifont->colortable[2] = highest_index;
			Machine->uifont->colortable[3] = lowest_index;
		}	
#endif
		// mark the whole screen dirty
		osd_mark_dirty(0, 0, Machine->scrbitmap->width-1,Machine->scrbitmap->height-1);
	}
}



void osd_get_pen(int pen,unsigned char *red, unsigned char *green, unsigned char *blue)
{
//	printf("osd_get_pen(%d,%d,%d,%d)\n", pen, red, green, blue);
	if (current_palette)
	{
		*red =	 current_palette[3*pen+0];
		*green = current_palette[3*pen+1];
		*blue =  current_palette[3*pen+2];
	}
	else
	{
		*red = 0;
		*green = 0;
		*blue = 0;
	}
}



static void update_screen_dummy(struct osd_bitmap *bitmap)
{
	logerror("dreamcast/video.c: undefined update_screen() function for %d x %d!\n",xmultiply,ymultiply);
}

INLINE void pan_display(void)
{
	int pan_changed = 0;

	/* horizontal panning */
	if (input_ui_pressed_repeat(IPT_UI_PAN_LEFT,1))
		if (skipcolumns < skipcolumnsmax)
		{
			skipcolumns++;
			osd_mark_dirty (0,0,Machine->scrbitmap->width-1,Machine->scrbitmap->height-1);
			pan_changed = 1;
		}
	if (input_ui_pressed_repeat(IPT_UI_PAN_RIGHT,1))
		if (skipcolumns > skipcolumnsmin)
		{
			skipcolumns--;
			osd_mark_dirty (0,0,Machine->scrbitmap->width-1,Machine->scrbitmap->height-1);
			pan_changed = 1;
		}
	if (input_ui_pressed_repeat(IPT_UI_PAN_DOWN,1))
		if (skiplines < skiplinesmax)
		{
			skiplines++;
			osd_mark_dirty (0,0,Machine->scrbitmap->width-1,Machine->scrbitmap->height-1);
			pan_changed = 1;
		}
	if (input_ui_pressed_repeat(IPT_UI_PAN_UP,1))
		if (skiplines > skiplinesmin)
		{
			skiplines--;
			osd_mark_dirty (0,0,Machine->scrbitmap->width-1,Machine->scrbitmap->height-1);
			pan_changed = 1;
		}

	if (pan_changed)
	{
		if (use_dirty) init_dirty(1);

/* Digita, I've discoved that it's actually one less pixel, probably a bug in the UI code */
		set_ui_visarea (skipcolumns, skiplines, skipcolumns+gfx_display_columns-1, skiplines+gfx_display_lines-2);
	}
}



int osd_skip_this_frame(void)
{
	static const int skiptable[FRAMESKIP_LEVELS][FRAMESKIP_LEVELS] =
	{
		{ 0,0,0,0,0,0,0,0,0,0,0,0 },
		{ 0,0,0,0,0,0,0,0,0,0,0,1 },
		{ 0,0,0,0,0,1,0,0,0,0,0,1 },
		{ 0,0,0,1,0,0,0,1,0,0,0,1 },
		{ 0,0,1,0,0,1,0,0,1,0,0,1 },
		{ 0,1,0,0,1,0,1,0,0,1,0,1 },
		{ 0,1,0,1,0,1,0,1,0,1,0,1 },
		{ 0,1,0,1,1,0,1,0,1,1,0,1 },
		{ 0,1,1,0,1,1,0,1,1,0,1,1 },
		{ 0,1,1,1,0,1,1,1,0,1,1,1 },
		{ 0,1,1,1,1,1,0,1,1,1,1,1 },
		{ 0,1,1,1,1,1,1,1,1,1,1,1 }
	};

	return skiptable[frameskip][frameskip_counter];
}

/* Update the display. */
void osd_update_video_and_audio(struct osd_bitmap *game_bitmap,struct osd_bitmap *debug_bitmap,int leds_status)
{
	static const int waittable[FRAMESKIP_LEVELS][FRAMESKIP_LEVELS] =
	{
		{ 1,1,1,1,1,1,1,1,1,1,1,1 },
		{ 2,1,1,1,1,1,1,1,1,1,1,0 },
		{ 2,1,1,1,1,0,2,1,1,1,1,0 },
		{ 2,1,1,0,2,1,1,0,2,1,1,0 },
		{ 2,1,0,2,1,0,2,1,0,2,1,0 },
		{ 2,0,2,1,0,2,0,2,1,0,2,0 },
		{ 2,0,2,0,2,0,2,0,2,0,2,0 },
		{ 2,0,2,0,0,3,0,2,0,0,3,0 },
		{ 3,0,0,3,0,0,3,0,0,3,0,0 },
		{ 4,0,0,0,4,0,0,0,4,0,0,0 },
		{ 6,0,0,0,0,0,6,0,0,0,0,0 },
		{12,0,0,0,0,0,0,0,0,0,0,0 }
	};
	int i;
	static int showfps,showfpstemp;
	TICKER curr;
	static TICKER prev_measure,this_frame_base,prev;
	static int speed = 100;
	static int vups,vfcount;
	int have_to_clear_bitmap = 0;
	int already_synced;
	struct osd_bitmap *bitmap;

//	printf("osd_update_video_and_audio(%d,%d,%d)\n", game_bitmap, debug_bitmap, leds_status);
	
	if (debug_bitmap && keyboard_pressed_memory(IPT_UI_TOGGLE_DEBUG))
	{
		osd_debugger_focus(show_debugger ^ 1);
	}

	if (debugger_focus_changed)
	{
		debugger_focus_changed = 0;

		if (show_debugger)
			internal_set_visible_area(0,debug_bitmap->width-1,0,debug_bitmap->height-1,1);
		else
			internal_set_visible_area(vis_min_x,vis_max_x,vis_min_y,vis_max_y,0);
	}

	if (show_debugger && debug_bitmap) bitmap = debug_bitmap;
	else bitmap = game_bitmap;

	if (warming_up)
	{
		/* first time through, initialize timer */
		prev_measure = ticker() - FRAMESKIP_LEVELS * TICKS_PER_SEC/video_fps;
		warming_up = 0;
	}

	if (frameskip_counter == 0)
		this_frame_base = prev_measure + FRAMESKIP_LEVELS * TICKS_PER_SEC/video_fps;

	if (throttle)
	{
		static TICKER last;

		/* if too much time has passed since last sound update, disable throttling */
		/* temporarily - we wouldn't be able to keep synch anyway. */
		curr = ticker();
		if ((curr - last) > 2*TICKS_PER_SEC / video_fps)
			throttle = 0;
		last = curr;

		already_synced = 1;

		throttle = 1;
	}
	else
		already_synced = 1;


	if (osd_skip_this_frame() == 0)
	{
		if (showfpstemp)
		{
			showfpstemp--;
			if (showfps == 0 && showfpstemp == 0)
			{
				have_to_clear_bitmap = 1;
			}
		}


		if (input_ui_pressed(IPT_UI_SHOW_FPS))
		{
			if (showfpstemp)
			{
				showfpstemp = 0;
				have_to_clear_bitmap = 1;
			}
			else
			{
				showfps ^= 1;
				if (showfps == 0)
				{
					have_to_clear_bitmap = 1;
				}
			}
		}


		/* now wait until it's time to update the screen */
		if (throttle)
		{
			profiler_mark(PROFILER_IDLE);
			{
				TICKER target;

				
				/* wait for video sync but use normal throttling */
//				if (wait_vsync)
//					vsync();

				curr = ticker();

//				if (already_synced == 0)
				{
			/* wait only if the audio update hasn't synced us already */

					target = this_frame_base +
							frameskip_counter * TICKS_PER_SEC/video_fps;

			if (curr - target < 0)
			{
				do
				{
							curr = ticker();
				} while (curr - target < 0);
			}
				}

			}
			profiler_mark(PROFILER_END);
		}
		else curr = ticker();

		
		/* for the FPS average calculation */
		if (++frames_displayed == FRAMES_TO_SKIP)
			start_time = curr;
		else
			end_time = curr;

		
		if (frameskip_counter == 0)
		{
			int divdr;

/* Digita, fix overflow in floating point */
			divdr = video_fps * (curr - prev_measure) / (10 * FRAMESKIP_LEVELS);
			speed = (TICKS_PER_SEC + divdr/2) * 10 / divdr;

			prev_measure = curr;
		}
				
		prev = curr;

		vfcount += waittable[frameskip][frameskip_counter];
		if (vfcount >= video_fps)
		{
			extern int vector_updates; /* avgdvg_go_w()'s per Mame frame, should be 1 */


			vfcount = 0;
			vups = vector_updates;
			vector_updates = 0;
		}

		if (!show_debugger && (showfps || showfpstemp))
		{
			int fps;
			char buf[30];
			int divdr;


			divdr = 100 * FRAMESKIP_LEVELS;
			fps = (video_fps * (FRAMESKIP_LEVELS - frameskip) * speed + (divdr / 2)) / divdr;
			sprintf(buf,"%s%2d%4d%%%4d/%d fps",autoframeskip?"auto":"fskp",frameskip,speed,fps,(int)(video_fps+0.5));
			ui_text(bitmap,buf,Machine->uiwidth-strlen(buf)*Machine->uifontwidth,0);
			if (vector_game)
			{
				sprintf(buf," %d vector updates",vups);
				ui_text(bitmap,buf,Machine->uiwidth-strlen(buf)*Machine->uifontwidth,Machine->uifontheight);
			}
		}

		{
			if (dirty_bright)
			{
				dirty_bright = 0;
				for (i = 0;i < 256;i++)
				{
					int r;
					
					r = i;
					
//					r = 63.0 * pow((float)(r) / 255.0, 1.0 / osd_gamma_correction);
					
					r = (int)(r * brightness * brightness_paused_adjust / 100.0);

					bright_lookup[i] = r;
				}
			}
			if (dirtypalette)
			{
				dirtypalette = 0;
				{
					for (i = 0;i < screen_colors;i++)
					{
						if (dirtycolor[i])
						{
							int r,g,b;
							
							dirtycolor[i] = 0;
							
							r = current_palette[3*i+0];
							g = current_palette[3*i+1];
							b = current_palette[3*i+2];
							
							r = bright_lookup[r];
							g = bright_lookup[g];
							b = bright_lookup[b];

							current_pens_lookup[i] = ((r >> 3) << 11) | ((g >> 2) << 5) | ((b >> 3) << 0);
						}
					}
				}
			}
		}

		if (show_debugger)
		{
			update_screen(bitmap);
        }
		else
		{
			/* copy the bitmap to screen memory */
			profiler_mark(PROFILER_BLIT);
			update_screen(bitmap);
			profiler_mark(PROFILER_END);
		}

		/* see if we need to give the card enough time to draw both odd/even fields of the interlaced display
			(req. for 15.75KHz Arcade Monitor Modes */
//		interlace_sync();


		if (!show_debugger && have_to_clear_bitmap)
			osd_clearbitmap(bitmap);

		if (use_dirty)
		{
			if (!vector_game)
				swap_dirty();
			init_dirty(0);
		}

		if (!show_debugger && have_to_clear_bitmap)
			osd_clearbitmap(bitmap);


		if (throttle && autoframeskip && frameskip_counter == 0)
		{
			static int frameskipadjust;
			int adjspeed;

			/* adjust speed to video refresh rate if vsync is on */
			adjspeed = speed * video_fps / vsync_frame_rate;

			if (adjspeed >= 100)
			{
				frameskipadjust++;
				if (frameskipadjust >= 3)
				{
					frameskipadjust = 0;
					if (frameskip > 0) frameskip--;
				}
			}
			else
			{
				if (adjspeed < 80)
					frameskipadjust -= (90 - adjspeed) / 5;
				else
				{
					/* don't push frameskip too far if we are close to 100% speed */
					if (frameskip < 8)
						frameskipadjust--;
				}

				while (frameskipadjust <= -2)
				{
					frameskipadjust += 2;
					/* Digita, we don't work so well with frameskip of 12 */
					if (frameskip < 8) frameskip++;
				}
			}
		}
	}

	/* Check for PGUP, PGDN and pan screen */
	pan_display();

	if (input_ui_pressed(IPT_UI_FRAMESKIP_INC))
	{
		if (autoframeskip)
		{
			autoframeskip = 0;
			frameskip = 0;
		}
		else
		{
			if (frameskip == FRAMESKIP_LEVELS-1)
			{
				frameskip = 0;
				autoframeskip = 1;
			}
			else
				frameskip++;
		}

		if (showfps == 0)
			showfpstemp = 2*video_fps;

		/* reset the frame counter every time the frameskip key is pressed, so */
		/* we'll measure the average FPS on a consistent status. */
		frames_displayed = 0;
	}

	if (input_ui_pressed(IPT_UI_FRAMESKIP_DEC))
	{
		if (autoframeskip)
		{
			autoframeskip = 0;
			frameskip = FRAMESKIP_LEVELS-1;
		}
		else
		{
			if (frameskip == 0)
				autoframeskip = 1;
			else
				frameskip--;
		}

		if (showfps == 0)
			showfpstemp = 2*video_fps;

		/* reset the frame counter every time the frameskip key is pressed, so */
		/* we'll measure the average FPS on a consistent status. */
		frames_displayed = 0;
	}

	if (input_ui_pressed(IPT_UI_THROTTLE))
	{
		throttle ^= 1;

		/* reset the frame counter every time the throttle key is pressed, so */
		/* we'll measure the average FPS on a consistent status. */
		frames_displayed = 0;
	}


	frameskip_counter = (frameskip_counter + 1) % FRAMESKIP_LEVELS;

//	poll_joysticks();
}



void osd_set_gamma(float _gamma)
{
	int i;

	osd_gamma_correction = _gamma;

	for (i = 0;i < screen_colors;i++)
		dirtycolor[i] = 1;
	dirtypalette = 1;
	dirty_bright = 1;
}

float osd_get_gamma(void)
{
	return osd_gamma_correction;
}

/* brightess = percentage 0-100% */
void osd_set_brightness(int _brightness)
{
	int i;

	brightness = _brightness;

	for (i = 0;i < screen_colors;i++)
		dirtycolor[i] = 1;
	dirtypalette = 1;
	dirty_bright = 1;
}

int osd_get_brightness(void)
{
	return brightness;
}


void osd_save_snapshot(struct osd_bitmap *bitmap)
{
	save_screen_snapshot(bitmap);
}

void osd_pause(int paused)
{
	int i;

	if (paused) brightness_paused_adjust = 0.65;
	else brightness_paused_adjust = 1.0;

	for (i = 0;i < screen_colors;i++)
		dirtycolor[i] = 1;
	dirtypalette = 1;
	dirty_bright = 1;
}

void osd_mark_vector_dirty(int x, int y)
{
	MARKDIRTY(y);
}

